#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2020-05-31
# @Filename: mwm_ob.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import peewee
from peewee import fn

from sdssdb.peewee.sdss5db.catalogdb import (Catalog, CatalogToTIC_v8,
                                             GAIA_ASSAS_SN_Cepheids, Gaia_DR2,
                                             Gaia_DR2_RUWE,
                                             Gaia_DR2_TwoMass_Best_Neighbour,
                                             TIC_v8, TwoMassPSC)

from . import BaseCarton


TMBN = Gaia_DR2_TwoMass_Best_Neighbour


class MWM_OB_Carton(BaseCarton):
    """Milky Way OB stars.

    Definition: Select all the hot, young stars in Gaia and 2MASS  with M_K < 0
    mag (M ~ 4 M_Sun) with Gaia G < 16 mag in the Milky Way, then subsampled
    to reduce the sample size / observing burden.

    Query on the Gaia archive:

        SELECT  g.source_id
        FROM gaiadr2.gaia_source as g
        JOIN gaiadr2.ruwe AS r
            USING (source_id)
        INNER JOIN gaiadr2.tmass_best_neighbour AS xmatch
            ON g.source_id = xmatch.source_id
        INNER JOIN gaiadr1.tmass_original_valid AS tm
            ON tm.tmass_oid = xmatch.tmass_oid
        WHERE parallax < power(10., (10. - tm.ks_m) / 5.)
            AND g.phot_g_mean_mag < 16.
            AND tm.j_m - tm.ks_m - 0.25 * (g.phot_g_mean_mag - tm.ks_m) < 0.10
            AND tm.j_m - tm.ks_m - 0.25 * (g.phot_g_mean_mag - tm.ks_m) > -0.30
            AND tm.j_m - tm.h_m < 0.15 * (g.phot_g_mean_mag - tm.ks_m) + 0.05
            AND tm.j_m - tm.h_m > 0.15 * (g.phot_g_mean_mag - tm.ks_m) - 0.15
            AND tm.j_m - tm.ks_m < 0.23 * (g.phot_g_mean_mag - tm.ks_m) + 0.03
            AND g.phot_g_mean_mag > 2 * (g.phot_g_mean_mag - tm.ks_m) + 3.0
            AND g.phot_g_mean_mag < 2 * (g.phot_g_mean_mag - tm.ks_m) + 11.
            AND xmatch.angular_distance < 1.
            AND r.ruwe < 1.4;

    Notes:
        - ks_m is the same as twomass_psc.h_m.

    """

    name = 'mwm_ob_core'
    mapper = 'MWM'
    category = 'science'
    cadence = 'mwm_ob_3x1'
    program = 'mwm_ob'
    priority = 2910

    def build_query(self, version_id, query_region=None):

        km = TwoMassPSC.k_m
        hm = TwoMassPSC.h_m
        jm = TwoMassPSC.j_m
        Gm = Gaia_DR2.phot_g_mean_mag

        query = (CatalogToTIC_v8
                 .select(CatalogToTIC_v8.catalogid,
                         Gaia_DR2.parallax,
                         Gaia_DR2.source_id.alias('gaia_source_id'),
                         km.alias('ks_m'))
                 .join(TIC_v8)
                 .join(Gaia_DR2)
                 .join(Gaia_DR2_RUWE,
                       on=(Gaia_DR2.source_id == Gaia_DR2_RUWE.source_id))
                 .join(TMBN,
                       on=(TMBN.source_id == Gaia_DR2_RUWE.source_id))
                 .join(TwoMassPSC,
                       on=(TMBN.tmass_pts_key == TwoMassPSC.pts_key))
                 .where(CatalogToTIC_v8.version_id == version_id,
                        CatalogToTIC_v8.best >> True)
                 .where(Gaia_DR2.parallax < fn.pow(10, ((10. - km - 0.61) / 5.)),
                        Gm < 16.,
                        Gm > 12,
                        jm - km - 0.25 * (Gm - km) < 0.10,
                        jm - km - 0.25 * (Gm - km) > -0.30,
                        jm - hm < 0.15 * (Gm - km) + 0.05,
                        jm - hm > 0.15 * (Gm - km) - 0.15,
                        jm - km < 0.23 * (Gm - km) + 0.03,
                        Gm > 2 * (Gm - km) + 3.0,
                        TMBN.angular_distance < 1.))

        if query_region:
            query = (query
                     .join_from(CatalogToTIC_v8, Catalog)
                     .where(peewee.fn.q3c_radial_query(Catalog.ra,
                                                       Catalog.dec,
                                                       query_region[0],
                                                       query_region[1],
                                                       query_region[2])))

        return query

    def post_process(self, Model):
        """Subsample of main sample.

        See files  https://keeper.mpdl.mpg.de/f/a84318bb5213431ea513/?dl=1 and
        https://keeper.mpdl.mpg.de/f/ee2d70d8eb2f463f9562/?dl=1.

        After the gamma-3 sims it was decided that the subsampling was not
        needed but we leave it here just in case.

        """

        return

        # data = numpy.array(Model.select(Model.catalogid,
        #                                 Model.ks_m,
        #                                 Model.parallax).tuples(),
        #                    dtype=[('catalogid', numpy.int64),
        #                           ('ks_m', float),
        #                           ('parallax', float)])

        # self.log.debug(f'Number of initial rows: {len(data)}.')

        # data = data[data['parallax'] > 0]
        # M_K_star = data['ks_m'] + 5. * numpy.log10(data['parallax'] / 100.)
        # data_sel_idx = numpy.isfinite(M_K_star)
        # data_sel = data[data_sel_idx]
        # M_K_star = M_K_star[data_sel_idx]

        # alpha = 0.4
        # M_K = numpy.arange(min(M_K_star), 0., 0.01)
        # p = erf(-alpha * (M_K - 0.1))

        # catalogid_new = []

        # # p ~ 3000 so this for loop is actually not that inefficient.
        # for i in range(len(p) - 1):

        #     w = numpy.where((M_K_star < M_K[i + 1]) & (M_K_star >= M_K[i]))
        #     p_ = numpy.mean(p[i:i + 1])

        #     N = len(data[w])
        #     N_new = numpy.int(numpy.round(p_ * N, 2))
        #     catalogid_ = numpy.random.choice(data_sel['catalogid'][w], N_new,
        #                                      replace=False)
        #     catalogid_new += catalogid_.tolist()

        # catalogid_new = set(catalogid_new)

        # self.log.debug(f'Number of selected rows: {len(catalogid_new)}.')

        # self.log.debug('Applying selected mask.')

        # values = ValuesList(zip(catalogid_new), columns=('catalogid',),
        #                     alias='vl')

        # with self.database.atomic():
        #     # Change everything to selected=False
        #     (Model.update({Model.selected: False}).execute())
        #     # Select the catalogids we calculated. If we tried to do a
        #     # .where(Model.catalogid != values.c.catalogid) that would take
        #     # forever, not sure why.
        #     (Model
        #      .update({Model.selected: True})
        #      .from_(values)
        #      .where(Model.catalogid == values.c.catalogid)
        #      .execute())


# class MWM_OB_MC_Carton(BaseCarton):
#     """Magellanic Clouds OB stars.

#     Definition: Select all the hot, young stars in Gaia and 2MASS with M_K < 0
#     mag (M ~ 4 M_Sun) with Gaia G < 16 mag in the Magellanic Clouds.

#     Pseudo-query:

#         phot_g_mean_ mag < 16.0
#         AND parallax < power(10., (10. -  ks_m) / 5.)
#         AND jm - ks_m - 0.25 * (phot_g_mean_mag - ks_m) < 0.10
#         AND j_m - h_m < 0.15 * (phot_g_mean_mag -ks_m) + 0.05
#         AND j_m - ks_m < 0.23 * (phot_g_mean_mag -ks_m) + 0.03
#         AND phot_g_mean_mag > 2 * (phot_g_mean_mag -ks_m) + 3.0
#         AND  265.0 < l < 310.0
#         AND -50 < b < -25.0
#         AND xmatch.angular distance < 1.0

#     Notes:
#         - ks_m is the same as twomass_psc.h_m.

#     """

#     name = 'mwm_ob_mc'
#     mapper = 'MWM'
#     category = 'science'
#     cadence = 'mwm_ob_3x1'
#     program = 'OB'

#     def build_query(self, version_id, query_region=None):

#         b = Gaia_DR2.b
#         l = Gaia_DR2.l  # noqa

#         km = TwoMassPSC.k_m
#         hm = TwoMassPSC.h_m
#         jm = TwoMassPSC.j_m
#         Gm = Gaia_DR2.phot_g_mean_mag

#         query = (CatalogToTIC_v8
#                  .select(CatalogToTIC_v8.catalogid,
#                          Gaia_DR2.source_id.alias('gaia_source_id'))
#                  .join(TIC_v8)
#                  .join(Gaia_DR2)
#                  .join(TMBN, on=(TMBN.source_id == Gaia_DR2.source_id))
#                  .join(TwoMassPSC,
#                        on=(TMBN.tmass_pts_key == TwoMassPSC.pts_key))
#                  .where(CatalogToTIC_v8.version_id == version_id,
#                         CatalogToTIC_v8.best >> True)
#                  .where(Gaia_DR2.parallax < fn.pow(10, ((10. - km) / 5.)),
#                         Gm < 16.,
#                         Gm > 12,
#                         jm - km - 0.25 * (Gm - km) < 0.10,
#                         jm - hm < 0.15 * (Gm - km) + 0.05,
#                         jm - km < 0.23 * (Gm - km) + 0.03,
#                         Gm > 2 * (Gm - km) + 3.0,
#                         TMBN.angular_distance < 1.,
#                         b > -50., b < -25., l > 265., l < 310.))

#         if query_region:
#             query = (query
#                      .join_from(CatalogToTIC_v8, Catalog)
#                      .where(peewee.fn.q3c_radial_query(Catalog.ra,
#                                                        Catalog.dec,
#                                                        query_region[0],
#                                                        query_region[1],
#                                                        query_region[2])))

#         return query


class MWM_OB_Cepheids_Carton(BaseCarton):
    """Milky Way Cepheids.

    Definition: List of Cepheids compiled by Inno et al. (in prep). The
    catalogue is obtained by using Gaia DR2 and ASAS-SN. Gaia parallaxes and
    variability information are used to select an initial sample, for which
    multi-epoch V-band photometry from the ASAS-SN survey is retrieved. From
    the ASAS-SN lightcurves, periods and Fourier parameters are derived. The
    Fourier parameters are used to identify classical Cepheids. The final
    catalogue is assembled by including several public classic Cepheid
    databases: OGLE-cCs, Gaia-cCs, ASASSN-cCs, VSX-cCs, Simbad-cCs,  and
    WISE-cCs. The final catalogue consists of ~3000 targets, between
    6 < G < 17 mag.

    """

    name = 'mwm_ob_cepheids'
    mapper = 'MWM'
    category = 'science'
    cadence = 'mwm_ob_3x1'
    program = 'mwm_ob'
    priority = 2910

    def build_query(self, version_id, query_region=None):

        query = (GAIA_ASSAS_SN_Cepheids
                 .select(CatalogToTIC_v8.catalogid,
                         Gaia_DR2.source_id)
                 .join(Gaia_DR2)
                 .join(TIC_v8)
                 .join(CatalogToTIC_v8)
                 .where(CatalogToTIC_v8.version_id == version_id,
                        CatalogToTIC_v8.best >> True,
                        Gaia_DR2.phot_g_mean_mag > 12))

        if query_region:
            query = (query
                     .join_from(CatalogToTIC_v8, Catalog)
                     .where(peewee.fn.q3c_radial_query(Catalog.ra,
                                                       Catalog.dec,
                                                       query_region[0],
                                                       query_region[1],
                                                       query_region[2])))

        return query



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2020-06-05
# @Filename: mwm_cb.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import peewee

from sdssdb.peewee.sdss5db.catalogdb import (CataclysmicVariables, Catalog,
                                             CatalogToGUVCat, CatalogToTIC_v8,
                                             Gaia_DR2,
                                             GeometricDistances_Gaia_DR2,
                                             GUVCat, TIC_v8)

from . import BaseCarton


class MWM_CB_300_Carton(BaseCarton):
    """MWM Compact Binaries 300pc.

    Definition: Cross-match Gaia & Bailer-Jones distances by source_id,
    cross-match with GALEX including proper motion corrections. Define a
    single linear relation in absolute FUV magnitude vs FUV - NUV,
    select all objects with distances less than 300pc.

    SQL:
        (FUVmag - 5 * log10(r_est/10)) < 14 * (FUVmag-NUVmag) - 46 & r_est <300

    Cadence: This target sample will be split into
        (1) H<11 to be observed with APOGEE.
        (2) phot_g_mean_mag < 16 to be observed with BOSS in bright time.
        (3) phot_g_mean_mag > 16 to be observed with BOSS in dark time.

    """

    name = 'mwm_cb_300pc'
    mapper = 'MWM'
    category = 'science'
    program = 'mwm_cb'
    cadence = 'mwm_cb_2x1'
    priority = 1400

    def build_query(self, version_id, query_region=None):

        GD = GeometricDistances_Gaia_DR2

        FUV = GUVCat.fuv_mag
        NUV = GUVCat.nuv_mag

        FUV_abs = FUV - 5 * peewee.fn.log(GD.r_est / 10)

        query = (CatalogToTIC_v8
                 .select(CatalogToTIC_v8.catalogid,
                         GD.source_id,
                         GD.r_est,
                         FUV, NUV,
                         TIC_v8.hmag.alias('h'))
                 .join(TIC_v8)
                 .join(Gaia_DR2)
                 .join(GD)
                 .join_from(CatalogToTIC_v8, CatalogToGUVCat,
                            on=(CatalogToGUVCat.catalogid == CatalogToTIC_v8.catalogid))
                 .join(GUVCat)
                 .where(GD.r_est < 300,
                        FUV_abs < 14 * (FUV - NUV) - 46)
                 .where(CatalogToGUVCat.version_id == version_id,
                        CatalogToGUVCat.best >> True,
                        CatalogToTIC_v8.version_id == version_id,
                        CatalogToTIC_v8.best >> True))

        if query_region:
            query = (query
                     .join_from(CatalogToTIC_v8, Catalog)
                     .where(peewee.fn.q3c_radial_query(Catalog.ra,
                                                       Catalog.dec,
                                                       query_region[0],
                                                       query_region[1],
                                                       query_region[2])))

        return query


class MWM_CB_Gaia_Galex_Carton(BaseCarton):
    """MWM Compact Binaries 300pc.

    Definition: Cross-match GALEX & Gaia, taking into account proper motions.

    SQL:
        parallax / parallax_error > 3 & gaiag < 20 &
        fuvmag + 5 * log10(parallax / 1000) + 5 > 1.5 + 1.28 * (fuvmag - gaiag)

    Cadence: This target sample will be split into
        (1) H<11 to be observed with APOGEE.
        (2) phot_g_mean_mag < 16 to be observed with BOSS in bright time.
        (3) phot_g_mean_mag > 16 to be observed with BOSS in dark time.

    """

    name = 'mwm_cb_gaiagalex'
    mapper = 'MWM'
    category = 'science'
    program = 'mwm_cb'
    cadence = 'mwm_cb_2x1'
    priority = 1400

    def build_query(self, version_id, query_region=None):

        FUV = GUVCat.fuv_mag
        gaiag = Gaia_DR2.phot_g_mean_mag

        FUV_abs = FUV + 5 * peewee.fn.log(Gaia_DR2.parallax / 1000.) + 5

        query = (CatalogToTIC_v8
                 .select(CatalogToTIC_v8.catalogid,
                         Gaia_DR2.source_id.alias('gaia_source_id'),
                         Gaia_DR2.parallax, Gaia_DR2.parallax_error,
                         FUV, gaiag.alias('phot_g_mean_mag'),
                         TIC_v8.hmag.alias('h'))
                 .join(TIC_v8)
                 .join(Gaia_DR2)
                 .join_from(CatalogToTIC_v8, CatalogToGUVCat,
                            on=(CatalogToGUVCat.catalogid == CatalogToTIC_v8.catalogid))
                 .join(GUVCat)
                 .where(Gaia_DR2.parallax / Gaia_DR2.parallax_error > 3,
                        gaiag < 20,
                        FUV > -999.,
                        FUV_abs > (1.5 + 1.28 * (FUV - gaiag)))
                 .where(CatalogToGUVCat.version_id == version_id,
                        CatalogToGUVCat.best >> True,
                        CatalogToTIC_v8.version_id == version_id,
                        CatalogToTIC_v8.best >> True))

        if query_region:
            query = (query
                     .join_from(CatalogToTIC_v8, Catalog)
                     .where(peewee.fn.q3c_radial_query(Catalog.ra,
                                                       Catalog.dec,
                                                       query_region[0],
                                                       query_region[1],
                                                       query_region[2])))

        return query


class MWM_CB_CV_Candidates_Carton(BaseCarton):
    """MWM Compact Binaries 300pc.

    Definition:
        List of cataclysmic variables (+candidates)
        compiled by the AAVSO.

    SQL:
        Select all records from table cataclysmic_variables.

    Cadence: This target sample will be split into
        (1) H<11 to be observed with APOGEE.
        (2) phot_g_mean_mag < 16 to be observed with BOSS in bright time.
        (3) phot_g_mean_mag > 16 to be observed with BOSS in dark time.

    """

    name = 'mwm_cb_cvcandidates'
    mapper = 'MWM'
    category = 'science'
    program = 'mwm_cb'
    cadence = 'mwm_cb_2x1'
    priority = 1400

    def build_query(self, version_id, query_region=None):

        query = (CatalogToTIC_v8
                 .select(CatalogToTIC_v8.catalogid,
                         CataclysmicVariables.source_id,
                         CataclysmicVariables.phot_g_mean_mag,
                         TIC_v8.hmag.alias('h'))
                 .join(TIC_v8)
                 .join(CataclysmicVariables,
                       on=(CataclysmicVariables.source_id == TIC_v8.gaia_int))
                 .where(CatalogToTIC_v8.version_id == version_id,
                        CatalogToTIC_v8.best >> True))

        if query_region:
            query = (query
                     .join_from(CatalogToTIC_v8, Catalog)
                     .where(peewee.fn.q3c_radial_query(Catalog.ra,
                                                       Catalog.dec,
                                                       query_region[0],
                                                       query_region[1],
                                                       query_region[2])))

        return query



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: Tom Dwelly
# @Date: 2020-06-12
# @Filename: bhm_gua.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

# flake8: noqa
# isort: skip_file

import peewee
import sdssdb
from peewee import JOIN

from sdssdb.peewee.sdss5db.catalogdb import (Catalog,
                                             CatalogToTIC_v8,
                                             TIC_v8,
                                             Gaia_DR2,
#this is old and does not work               CatalogToGaia_unWISE_AGN,
                                             CatalogToSDSS_DR16_SpecObj,
                                             SDSS_DR16_SpecObj,
                                             Gaia_unWISE_AGN)



from target_selection.cartons.base import BaseCarton


# Details: Start here
# https://wiki.sdss.org/display/OPS/Defining+target+selection+and+cadence+algorithms

## This module provides the following BHM cartons:
# bhm_gua_dark
# bhm_gua_bright


'''
  pseudo-SQL

    both Cartons:
        SELECT * FROM gaia_unwise_agn AS gua
        LEFT JOIN sdss_dr16_specobj AS so ON  q3c_join(gua.ra, gua.dec, so.plug_ra, so.plug_dec, 1.0)
        AND WHERE gua.prob_rf > 0.8
        AND WHERE (so.specobjid = NULL OR so.zwarning != 0 OR so.sn_median_all < x.x OR so.z_err > 0.0xx )

    bhm_gua_dark
        AND WHERE ( gua.g > 16.5 AND gua.rp > 16.5 AND (gua.g < 21.2 OR gua.rp < 21.0 ) )

    bhm_gua_bright
        AND WHERE ( gua.g > 13.0 AND gua.rp > 13.5 AND (gua.g < 18.5 OR gua.rp < 18.5) )
'''



class BhmGuaBaseCarton(BaseCarton):
    '''
    Parent class that provides the basic selections for both Gaia UnWISE AGN cartons
    To be sub-classed, not to be called directly.

    To get from Catalog to GUA we join tables via :
    Catalog -> CatalogToTIC_v8 -> Gaia_DR2 -> Gaia_unWISE_AGN
    '''

    name = 'bhm_gua_base'
    category = 'science'
    mapper = 'BHM'
    program = 'bhm_filler'
    tile = False
    priority = None

    def build_query(self, version_id, query_region=None):
        c = Catalog.alias()
        ##############c2t = CatalogToGaia_unWISE_AGN.alias() - deprecated
        c2tic = CatalogToTIC_v8.alias()
        tic = TIC_v8.alias()
        c2s = CatalogToSDSS_DR16_SpecObj.alias()
        g = Gaia_DR2.alias()
        t = Gaia_unWISE_AGN.alias()
        s = SDSS_DR16_SpecObj.alias()

        # set the Carton priority+values here - read from yaml
        priority = peewee.Value(int(self.parameters.get('priority', 10000))).alias('priority')
        value = peewee.Value(self.parameters.get('value', 1.0)).cast('float').alias('value')
        pmra = peewee.Value(0.0).cast('float').alias('pmra')
        pmdec = peewee.Value(0.0).cast('float').alias('pmdec')
        parallax = peewee.Value(0.0).cast('float').alias('parallax')

        query = (
            c.select(c.catalogid,
                     priority,
                     value,
                     pmra,
                     pmdec,
                     parallax,
#                     t.g.alias('g'),   ##rely on the centralised magnitude routines
#                     t.bp.alias('bp'),
#                     t.rp.alias('rp'),
            )
            .join(c2tic)
            .join(tic)
            .join(g)
            .join(t, on=(g.source_id == t.gaia_sourceid))
            .where(c.version_id == version_id,
                   c2tic.version_id == version_id,
                   c2tic.best == True)
            .where(
                (t.prob_rf >= self.parameters['prob_rf_min']),
                (t.g >= self.parameters['mag_g_min']),
                (t.rp >= self.parameters['mag_rp_min']),
                (
                    (t.g < self.parameters['mag_g_max']) |
                    (t.rp < self.parameters['mag_rp_max'])
                ),
            )
            .switch(c)
            .join(c2s, JOIN.LEFT_OUTER)
            .join(s, JOIN.LEFT_OUTER)
            .where(
                (s.specobjid.is_null()) |
                (s.zwarning != 0 ) |
                (s.snmedian < self.parameters['spec_sn_thresh']) |
                (s.zerr >  self.parameters['spec_z_err_thresh'])
            )
            .distinct([t.gaia_sourceid])   # avoid duplicates - trust the GUA parent sample
        )

        return query


#-------  bhm_gua_dark   ------ #

class BhmGuaDarkCarton(BhmGuaBaseCarton):
    '''
        AND WHERE ( gua.g > 16.x AND gua.rp > 16.x)
    '''
    name = 'bhm_gua_dark'
    cadence = 'bhm_spiders_1x4'

#-------  bhm_gua_bright   ------ #

class BhmGuaBrightCarton(BhmGuaBaseCarton):
    '''
        AND WHERE ( gua.g < 18.x OR gua.rp < 18.x)
    '''
    name = 'bhm_gua_bright'
    cadence = 'bhm_boss_bright_3x1'




'''
Exporting from the temp table

\copy (SELECT * FROM sandbox.temp_bhm_gaia_unwise_agn_dark)  TO '/home/tdwelly/scratch/targetdb/bhm_gaia_unwise_agn_dark.csv' with csv header
\copy (SELECT * FROM sandbox.temp_bhm_gaia_unwise_agn_bright)  TO '/home/tdwelly/scratch/targetdb/bhm_gaia_unwise_agn_bright.csv' with csv header

for F in bhm_gaia_unwise_agn_*.csv; do   stilts tpipe in=${F} out="${F%.*}.fits" ifmt=csv ofmt=fits-basic; done
m
'''



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: Pramod Gupta (psgupta@uw.edu)
# @Date: 2020-06-10
# @Filename: mwm_yso.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import peewee

from sdssdb.peewee.sdss5db.catalogdb import (MIPSGAL, AllWise, Catalog,
                                             CatalogToTIC_v8, Gaia_DR2,
                                             Sagitta, TIC_v8, TwoMassPSC,
                                             YSO_Clustering, Zari18pms)

from target_selection.cartons import BaseCarton


# See catalog.py for the name of peewee model names corresponding
# to postgres table names:
# https://github.com/sdss/sdssdb/blob/master/python/sdssdb/peewee/sdss5db/catalogdb.py

class MWM_YSO_Disk_APOGEE_Carton(BaseCarton):
    """YSOs - Disk APOGEE (IR excess).

    Shorthand name: mwm_yso_disk_apogee

    old class name: MWM_YSO_S1_Carton
    old shorthand name: mwm_yso_s1

    Simplified Description of selection criteria:
    selection of YSOs based on IR excess,
    with WISE colors W1-W2>0.25, W2-W3>0.5, W3-W4>1.5,
    closer than parallax>0.3, and brighter than H<13
    (should have ~21.5K sources)
    Wiki page:
    https://wiki.sdss.org/display/MWM/YSO+selection+function
    Additional source catalogs needed: Gaia, 2mass, allwise
    Additional cross-matching needed:
    Note: Using the Gaia xmatch somehow misses half the sources.
    Selection was done on the allwise catalog that
    had 2mass photometry,
    and then the resulting selection was crossmatched against against
    Gaia with 1" search radius.
    Return columns: Gaia id, 2mass id, allwise id, G, BP, RP,
    J, H, K, W1, W2, W3, W4,parallax
    cadence options for these targets
    (list all options,
    even though no single target will receive more than one):
    Pseudo SQL (optional):
    Implementation: h_m<13 and w1mpro-w2mpro>0.25 and
    w2mpro-w3mpro>0.5 and w3mpro-w4mpro>1.5 and parallax>0.3

    """

    name = 'mwm_yso_disk_apogee'
    category = 'science'
    cadence = None  # 'apogee_bright_3x1'
    program = 'mwm_yso'
    mapper = 'MWM'
    priority = 2700

    def build_query(self, version_id, query_region=None):

        query = (CatalogToTIC_v8
                 .select(CatalogToTIC_v8.catalogid, Gaia_DR2.source_id,
                         Gaia_DR2.ra.alias('gaia_dr2_ra'),
                         Gaia_DR2.dec.alias('gaia_dr2_dec'),
                         TwoMassPSC.pts_key,
                         TwoMassPSC.designation.alias('twomass_psc_designation'),
                         AllWise.designation.alias('allwise_designation'),
                         Gaia_DR2.phot_g_mean_mag, Gaia_DR2.phot_bp_mean_mag,
                         Gaia_DR2.phot_rp_mean_mag.alias('gaia_dr2_rp'),
                         TwoMassPSC.j_m, TwoMassPSC.h_m,
                         TwoMassPSC.k_m,
                         Gaia_DR2.parallax)
                 .join(TIC_v8, on=(CatalogToTIC_v8.target_id == TIC_v8.id))
                 .join(Gaia_DR2, on=(TIC_v8.gaia_int == Gaia_DR2.source_id))
                 .switch(TIC_v8)
                 .join(TwoMassPSC, on=(TIC_v8.twomass_psc == TwoMassPSC.designation))
                 .switch(TIC_v8)
                 .join(AllWise, on=(TIC_v8.allwise == AllWise.designation))
                 .where(CatalogToTIC_v8.version_id == version_id,
                        CatalogToTIC_v8.best >> True,
                        TwoMassPSC.h_m < 13,
                        (AllWise.w1mpro - AllWise.w2mpro) > 0.25,
                        (AllWise.w2mpro - AllWise.w3mpro) > 0.50,
                        (AllWise.w3mpro - AllWise.w4mpro) > 1.50,
                        Gaia_DR2.parallax > 0.3))

        # Gaia_DR2 pweewee model class corresponds to
        # table catalogdb.gaia_dr2_source.
        #
        # All values of TIC_v8.plx (for non-null entries) are not the same as
        # values of Gaia_DR2.parallax.
        # Hence, in the above query, we cannot use TIC_v8.plx instead
        # of Gaia_DR2.parallax.

        if query_region:
            query = (query
                     .join_from(CatalogToTIC_v8, Catalog)
                     .where(peewee.fn.q3c_radial_query(Catalog.ra,
                                                       Catalog.dec,
                                                       query_region[0],
                                                       query_region[1],
                                                       query_region[2])))

        return query


class MWM_YSO_Disk_BOSS_Carton(BaseCarton):
    """YSOs - Disk BOSS (IR excess).

    Shorthand name: mwm_yso_disk_boss

    old class name: MWM_YSO_S1_Carton
    old shorthand name: mwm_yso_s1

    Simplified Description of selection criteria:
    selection of YSOs based on IR excess,
    with WISE colors W1-W2>0.25, W2-W3>0.5, W3-W4>1.5,
    closer than parallax>0.3, and brighter than H<13
    (should have ~21.5K sources)
    Wiki page:
    https://wiki.sdss.org/display/MWM/YSO+selection+function
    Additional source catalogs needed: Gaia, 2mass, allwise
    Additional cross-matching needed:
    Note: Using the Gaia xmatch somehow misses half the sources.
    Selection was done on the allwise catalog that
    had 2mass photometry,
    and then the resulting selection was crossmatched against against
    Gaia with 1" search radius.
    Return columns: Gaia id, 2mass id, allwise id, G, BP, RP,
    J, H, K, W1, W2, W3, W4,parallax
    cadence options for these targets
    (list all options,
    even though no single target will receive more than one):
    boss_bright_3x1 if RP<14.76 |
    boss_bright_4x1 if RP<15.075 |
    boss_bright_5x1 if RP<15.29 |
    boss_bright_6x1 if RP<15.5
    Pseudo SQL (optional):
    Implementation: phot_rp_mean_mag<15.5 and w1mpro-w2mpro>0.25 and
    w2mpro-w3mpro>0.5 and w3mpro-w4mpro>1.5 and parallax>0.3

    Comments: Split from mwm_yso_s1 to request BOSS observations,
    same color selection but assigning cadence and faint limit for carton based
    on RP instead of H
    """

    name = 'mwm_yso_disk_boss'
    category = 'science'
    # cadence is set in post_process()
    cadence = None
    program = 'mwm_yso'
    mapper = 'MWM'
    priority = 2700

    def build_query(self, version_id, query_region=None):

        query = (CatalogToTIC_v8
                 .select(CatalogToTIC_v8.catalogid, Gaia_DR2.source_id,
                         Gaia_DR2.ra.alias('gaia_dr2_ra'),
                         Gaia_DR2.dec.alias('gaia_dr2_dec'),
                         TwoMassPSC.pts_key,
                         TwoMassPSC.designation.alias('twomass_psc_designation'),
                         AllWise.designation.alias('allwise_designation'),
                         Gaia_DR2.phot_g_mean_mag, Gaia_DR2.phot_bp_mean_mag,
                         Gaia_DR2.phot_rp_mean_mag.alias('gaia_dr2_rp'),
                         TwoMassPSC.j_m, TwoMassPSC.h_m,
                         TwoMassPSC.k_m,
                         Gaia_DR2.parallax)
                 .join(TIC_v8, on=(CatalogToTIC_v8.target_id == TIC_v8.id))
                 .join(Gaia_DR2, on=(TIC_v8.gaia_int == Gaia_DR2.source_id))
                 .switch(TIC_v8)
                 .join(TwoMassPSC, on=(TIC_v8.twomass_psc == TwoMassPSC.designation))
                 .switch(TIC_v8)
                 .join(AllWise, on=(TIC_v8.allwise == AllWise.designation))
                 .where(CatalogToTIC_v8.version_id == version_id,
                        CatalogToTIC_v8.best >> True,
                        Gaia_DR2.phot_rp_mean_mag < 15.5,
                        (AllWise.w1mpro - AllWise.w2mpro) > 0.25,
                        (AllWise.w2mpro - AllWise.w3mpro) > 0.50,
                        (AllWise.w3mpro - AllWise.w4mpro) > 1.50,
                        Gaia_DR2.parallax > 0.3))

        # Gaia_DR2 pweewee model class corresponds to
        # table catalogdb.gaia_dr2_source.
        #
        # All values of TIC_v8.plx (for non-null entries) are not the same as
        # values of Gaia_DR2.parallax.
        # Hence, in the above query, we cannot use TIC_v8.plx instead
        # of Gaia_DR2.parallax.

        if query_region:
            query = (query
                     .join_from(CatalogToTIC_v8, Catalog)
                     .where(peewee.fn.q3c_radial_query(Catalog.ra,
                                                       Catalog.dec,
                                                       query_region[0],
                                                       query_region[1],
                                                       query_region[2])))

        return query

    def post_process(self, model):
        """
        cadence options for these targets:
        boss_bright_3x1 if RP<14.76 |
        boss_bright_4x1 if RP<15.075 |
        boss_bright_5x1 if RP<15.29 |
        boss_bright_6x1 if RP<15.5
        """

        cursor = self.database.execute_sql(
            "select catalogid, gaia_dr2_rp from " +
            " sandbox.temp_mwm_yso_disk_boss ;")

        output = cursor.fetchall()

        for i in range(len(output)):
            current_catalogid = output[i][0]
            current_rp = output[i][1]

            if(current_rp < 14.76):
                current_cadence = 'boss_bright_3x1'
            elif(current_rp < 15.075):
                current_cadence = 'boss_bright_4x1'
            elif(current_rp < 15.29):
                current_cadence = 'boss_bright_5x1'
            elif(current_rp < 15.5):
                current_cadence = 'boss_bright_6x1'
            else:
                current_cadence = None

            self.database.execute_sql(
                " update sandbox.temp_mwm_yso_disk_boss " +
                " set cadence = '" + current_cadence + "'"
                " where catalogid = " + str(current_catalogid) + ";")


class MWM_YSO_Embedded_APOGEE_Carton(BaseCarton):
    """YSOs - Embedded APOGEE (optically invisible).

    Shorthand name: mwm_yso_embedded_apogee

    old class name: MWM_YSO_S2_Carton
    old shorthand name: mwm_yso_s2

    Simplified Description of selection criteria:
    selection of YSOs, brighter than H<13, fainter than G>15 or
    without gaia detection,
    colors J-H>0,5, W1-W2>0.5, W2-W3>1, W3-W4>1.5, and
     relates (W3-W4)>(W1-W2)*0.5+1.1
    (should have ~11.6K sources)
    Wiki page:
    https://wiki.sdss.org/display/MWM/YSO+selection+function
    Additional source catalogs needed: 2mass+allwise, gaia
    (allow sources that lack gaia xmatch)
    Additional cross-matching needed:
    Note: Using the Gaia xmatch somehow misses half the sources.
    Selection was done on the allwise catalog
    that had 2mass photometry,
    and then the resulting selection was crossmatched
    against against Gaia with 1" search radius.
    Return columns: Gaia id, 2mass id, allwise id, G, BP, RP,
    J, H, K, W1, W2, W3, W4
    cadence options for these targets
    (list all options,
    even though no single target will receive more than one):
    Pseudo SQL (optional):
    Implementation: h_m<13 and
    (phot_g_mean_mag>18.5 or phot_g_mean_mag is null)
    and j_m-h_m>1
    and h_m-ks_m>0.5
    and w1mpro-w2mpro>0.5
    and w2mpro-w3mpro>1
    and w3mpro-w4mpro>1.5
    and w3mpro-w4mpro>(w1mpro-w2mpro)*0.8+1.1

    """

    name = 'mwm_yso_embedded_apogee'
    category = 'science'
    cadence = None  # 'apogee_bright_3x1'
    program = 'mwm_yso'
    mapper = 'MWM'
    priority = 2700

    def build_query(self, version_id, query_region=None):

        query = (AllWise
                 .select(CatalogToTIC_v8.catalogid, Gaia_DR2.source_id,
                         Gaia_DR2.ra.alias('gaia_dr2_ra'),
                         Gaia_DR2.dec.alias('gaia_dr2_dec'),
                         TwoMassPSC.pts_key,
                         TwoMassPSC.designation.alias('twomass_psc_designation'),
                         AllWise.designation.alias('allwise_designation'),
                         Gaia_DR2.phot_g_mean_mag, Gaia_DR2.phot_bp_mean_mag,
                         Gaia_DR2.phot_rp_mean_mag.alias('gaia_dr2_rp'),
                         TwoMassPSC.j_m, TwoMassPSC.h_m,
                         TwoMassPSC.k_m,
                         Gaia_DR2.parallax)
                 .join(TIC_v8, on=(TIC_v8.allwise == AllWise.designation))
                 .join(TwoMassPSC,
                       on=(TIC_v8.twomass_psc == TwoMassPSC.designation))
                 .switch(TIC_v8)
                 .join(Gaia_DR2, peewee.JOIN.LEFT_OUTER,
                       on=(TIC_v8.gaia_int == Gaia_DR2.source_id))
                 .switch(TIC_v8)
                 .join(CatalogToTIC_v8,
                       on=(CatalogToTIC_v8.target_id == TIC_v8.id))
                 .where(CatalogToTIC_v8.version_id == version_id,
                        CatalogToTIC_v8.best >> True,
                        TwoMassPSC.h_m < 13,
                        (Gaia_DR2.phot_g_mean_mag > 18.5) |
                        (Gaia_DR2.phot_g_mean_mag >> None),
                        (AllWise.j_m_2mass - AllWise.h_m_2mass) > 1.0,
                        (AllWise.h_m_2mass - AllWise.k_m_2mass) > 0.5,
                        (AllWise.w1mpro - AllWise.w2mpro) > 0.50,
                        (AllWise.w2mpro - AllWise.w3mpro) > 1.00,
                        (AllWise.w3mpro - AllWise.w4mpro) > 1.50,
                        (AllWise.w3mpro - AllWise.w4mpro) >
                        (AllWise.w1mpro - AllWise.w2mpro) * 0.8 + 1.1))

        if query_region:
            query = (query
                     .join_from(CatalogToTIC_v8, Catalog)
                     .where(peewee.fn.q3c_radial_query(Catalog.ra,
                                                       Catalog.dec,
                                                       query_region[0],
                                                       query_region[1],
                                                       query_region[2])))
        return query


class MWM_YSO_Nebula_APOGEE_Carton(BaseCarton):
    """YSOs - Nebula APOGEE(optically invisible, WISE saturated).

    Shorthand name: mwm_yso_nebula_apogee

    old class name: MWM_YSO_S2_5_Carton
    old shorthand name: mwm_yso_s2_5

    Simplified Description of selection criteria:
    selection of YSOs, brighter than H<15,
    saturated (blank) W4 with W2-W3>4,
    or saturated W3 and W2, with J-H>1.1.
    Some contaminants from scanning are
    filtered on the plane of the sky:
    all the targets should be within 5 deg of the plane+
    few sources that can be located
    further south of the plane if l>180
    (should have ~1.2K sources)
    Wiki page:
    https://wiki.sdss.org/display/MWM/YSO+selection+function
    Additional source catalogs needed: 2mass, allwise
    Additional cross-matching needed:
    Return columns: 2mass id, allwise id, J, H, K, W1, W2, W3, W4
    cadence options for these targets
    (list all options,
    even though no single target will receive more than one):
    Pseudo SQL (optional):
    Implementation: h_m<13 and
    (w2mpro-w3mpro>4 and w4mpro is null) or
    (w3mpro is null and w4mpro is null and j_m-h_m>1.1)
    and (b>-5 or l>180) and b<-5

    """

    name = 'mwm_yso_nebula_apogee'
    category = 'science'
    cadence = None  # 'apogee_bright_3x1'
    program = 'mwm_yso'
    mapper = 'MWM'
    priority = 2700

    # Above implementation has below clause
    # and (b>-5 or l>180) and b<-5
    # Replace (b>-5 or l>180) and b<-5 as below based on the text.
    # In words:
    # all the targets should be within 5 deg of the plane+
    # few sources that can be
    # located further south of the plane if l>180
    # Hence:
    # ((b>-5) and (b<5)) or ((b<-5) and (l > 180))
    #  l, b in Gaia_DR2 are gallong and gallat in TIC_v8.
    # We are using the values from Gaia since
    # TIC propagates the coordinates back to epoch 2000.0
    # (b>-5 or l>180) and b<-5
    # S2_5 query below has the same part before where() as S2 query.

    def build_query(self, version_id, query_region=None):

        query = (AllWise
                 .select(CatalogToTIC_v8.catalogid, Gaia_DR2.source_id,
                         Gaia_DR2.ra.alias('gaia_dr2_ra'),
                         Gaia_DR2.dec.alias('gaia_dr2_dec'),
                         TwoMassPSC.pts_key,
                         TwoMassPSC.designation.alias('twomass_psc_designation'),
                         AllWise.designation.alias('allwise_designation'),
                         Gaia_DR2.phot_g_mean_mag, Gaia_DR2.phot_bp_mean_mag,
                         Gaia_DR2.phot_rp_mean_mag.alias('gaia_dr2_rp'),
                         TwoMassPSC.j_m, TwoMassPSC.h_m,
                         TwoMassPSC.k_m,
                         Gaia_DR2.parallax)
                 .join(TIC_v8, on=(TIC_v8.allwise == AllWise.designation))
                 .join(TwoMassPSC,
                       on=(TIC_v8.twomass_psc == TwoMassPSC.designation))
                 .switch(TIC_v8)
                 .join(Gaia_DR2, peewee.JOIN.LEFT_OUTER,
                       on=(TIC_v8.gaia_int == Gaia_DR2.source_id))
                 .switch(TIC_v8)
                 .join(CatalogToTIC_v8,
                       on=(CatalogToTIC_v8.target_id == TIC_v8.id))
                 .where(CatalogToTIC_v8.version_id == version_id,
                        CatalogToTIC_v8.best >> True,
                        TwoMassPSC.h_m < 13,
                        (((AllWise.w2mpro - AllWise.w3mpro) > 4) &
                         (AllWise.w4mpro >> None)) |
                        ((AllWise.w3mpro >> None) &
                         (AllWise.w4mpro >> None) &
                         ((AllWise.j_m_2mass - AllWise.h_m_2mass) > 1.1)),
                        ((Gaia_DR2.b > -5) & (Gaia_DR2.b < 5)) |
                        ((Gaia_DR2.b < -5) & (Gaia_DR2.l > 180)) |
                        ((Gaia_DR2.b >> None) & (Gaia_DR2.l >> None))))

        if query_region:
            query = (query
                     .join_from(CatalogToTIC_v8, Catalog)
                     .where(peewee.fn.q3c_radial_query(Catalog.ra,
                                                       Catalog.dec,
                                                       query_region[0],
                                                       query_region[1],
                                                       query_region[2])))
        return query


class MWM_YSO_Variable_APOGEE_Carton(BaseCarton):
    """YSOs - Variable APOGEE (pre-main sequence optical variables).

    Shorthand name: mwm_yso_variable_apogee

    old class name: MWM_YSO_S3_Carton
    old shorthand name: mwm_yso_s3

    Simplified Description of selection criteria:
    selection of YSOs brighter than H<13, closer than parallax>0.3.
    Filter on the position of the HR diagram to
    select cool pre-main sequence stars,
    with BP-RP>13, (BP-RP)*2.5+2.5>M_G, (BP-RP)*2.5-1<M_G,
    requiring variability in g,bp,rp>0.02
    (with var_x defined as
    sqrt(phot_x_n_obs)/phot_x_mean_flux_over_error),
    have relations in variability of
    var_g<var_bp<var_g^0.75, 0.75*var_g<var_rp<var_g^0.95,
    and log10(var_bp)*5+11<M_BP, in which M_x is the absolute mag
    (should have ~52.7K sources)
    Wiki page:
    https://wiki.sdss.org/display/MWM/YSO+selection+function
    Additional source catalogs needed: 2mass, gaia
    Additional cross-matching needed:
    Return columns: Gaia id, 2mass id, G, BP, RP, J, H, K, parallax
    cadence options for these targets
    (list all options,
    even though no single target will receive more than one):
    Pseudo SQL (optional):
    Implementation:
    phot_g_mean_mag < 18.5 and h_m <13 and parallax >0.3 and
    bp_rp*2.5+2.5 > phot_g_mean_mag-5*(log10(1000/parallax)-1) and
    bp_rp*2.5-1 < phot_g_mean_mag-5*(log10(1000/parallax)-1) and
    sqrt(phot_bp_n_obs)/phot_bp_mean_flux_over_error>
    sqrt(phot_g_n_obs)/phot_g_mean_flux_over_error and
    sqrt(phot_rp_n_obs)/phot_rp_mean_flux_over_error>
    sqrt(phot_g_n_obs)/phot_g_mean_flux_over_error*0.75 and
    sqrt(phot_bp_n_obs)/phot_bp_mean_flux_over_error<
    power(sqrt(phot_g_n_obs)/phot_g_mean_flux_over_error,0.75) and
    sqrt(phot_rp_n_obs)/phot_rp_mean_flux_over_error<
    power(sqrt(phot_g_n_obs)/phot_g_mean_flux_over_error,0.95) and
    log10(sqrt(phot_bp_n_obs)/phot_bp_mean_flux_over_error)*5+11<
    phot_bp_mean_mag-5*(log10(1000/parallax)-1) and
    bp_rp>1.3 and sqrt(phot_g_n_obs)/phot_g_mean_flux_over_error>0.02
    and
    sqrt(phot_bp_n_obs)/phot_bp_mean_flux_over_error>0.02 and
    sqrt(phot_rp_n_obs)/phot_rp_mean_flux_over_error>0.02

    """

    name = 'mwm_yso_variable_apogee'
    category = 'science'
    cadence = None  # 'apogee_bright_3x1'
    program = 'mwm_yso'
    mapper = 'MWM'
    priority = 2700

    def build_query(self, version_id, query_region=None):

        query = (CatalogToTIC_v8
                 .select(CatalogToTIC_v8.catalogid, Gaia_DR2.source_id,
                         Gaia_DR2.ra.alias('gaia_dr2_ra'),
                         Gaia_DR2.dec.alias('gaia_dr2_dec'),
                         TwoMassPSC.pts_key,
                         TwoMassPSC.designation.alias('twomass_psc_designation'),
                         Gaia_DR2.phot_g_mean_mag, Gaia_DR2.phot_bp_mean_mag,
                         Gaia_DR2.phot_rp_mean_mag.alias('gaia_dr2_rp'),
                         TwoMassPSC.j_m, TwoMassPSC.h_m,
                         TwoMassPSC.k_m,
                         Gaia_DR2.parallax)
                 .join(TIC_v8, on=(CatalogToTIC_v8.target_id == TIC_v8.id))
                 .join(TwoMassPSC, on=(TIC_v8.twomass_psc == TwoMassPSC.designation))
                 .switch(TIC_v8)
                 .join(Gaia_DR2, on=(TIC_v8.gaia_int == Gaia_DR2.source_id))
                 .where(CatalogToTIC_v8.version_id == version_id,
                        CatalogToTIC_v8.best >> True,
                        Gaia_DR2.phot_g_mean_mag < 18.5,
                        TwoMassPSC.h_m < 13,
                        Gaia_DR2.parallax > 0.3,
                        Gaia_DR2.bp_rp * 2.5 + 2.5 >
                        Gaia_DR2.phot_g_mean_mag -
                        5 * (peewee.fn.log(1000 / Gaia_DR2.parallax) - 1),
                        Gaia_DR2.bp_rp * 2.5 - 1 <
                        Gaia_DR2.phot_g_mean_mag -
                        5 * (peewee.fn.log(1000 / Gaia_DR2.parallax) - 1),
                        peewee.fn.sqrt(Gaia_DR2.phot_bp_n_obs) /
                        Gaia_DR2.phot_bp_mean_flux_over_error >
                        peewee.fn.sqrt(Gaia_DR2.phot_g_n_obs) /
                        Gaia_DR2.phot_g_mean_flux_over_error,
                        peewee.fn.sqrt(Gaia_DR2.phot_rp_n_obs) /
                        Gaia_DR2.phot_rp_mean_flux_over_error >
                        peewee.fn.sqrt(Gaia_DR2.phot_g_n_obs) /
                        Gaia_DR2.phot_g_mean_flux_over_error * 0.75,
                        peewee.fn.sqrt(Gaia_DR2.phot_bp_n_obs) /
                        Gaia_DR2.phot_bp_mean_flux_over_error <
                        peewee.fn.power(
                            peewee.fn.sqrt(Gaia_DR2.phot_g_n_obs) /
                            Gaia_DR2.phot_g_mean_flux_over_error, 0.75),
                        peewee.fn.sqrt(Gaia_DR2.phot_rp_n_obs) /
                        Gaia_DR2.phot_rp_mean_flux_over_error <
                        peewee.fn.power(
                            peewee.fn.sqrt(Gaia_DR2.phot_g_n_obs) /
                            Gaia_DR2.phot_g_mean_flux_over_error, 0.95),
                        peewee.fn.log(
                            peewee.fn.sqrt(Gaia_DR2.phot_bp_n_obs) /
                            Gaia_DR2.phot_bp_mean_flux_over_error) * 5 + 11 <
                        Gaia_DR2.phot_bp_mean_mag -
                        5 * (peewee.fn.log(1000 / Gaia_DR2.parallax) - 1),
                        Gaia_DR2.bp_rp > 1.3,
                        peewee.fn.sqrt(Gaia_DR2.phot_g_n_obs) /
                        Gaia_DR2.phot_g_mean_flux_over_error > 0.02,
                        peewee.fn.sqrt(Gaia_DR2.phot_bp_n_obs) /
                        Gaia_DR2.phot_bp_mean_flux_over_error > 0.02,
                        peewee.fn.sqrt(Gaia_DR2.phot_rp_n_obs) /
                        Gaia_DR2.phot_rp_mean_flux_over_error > 0.02))

        if query_region:
            query = (query
                     .join_from(CatalogToTIC_v8, Catalog)
                     .where(peewee.fn.q3c_radial_query(Catalog.ra,
                                                       Catalog.dec,
                                                       query_region[0],
                                                       query_region[1],
                                                       query_region[2])))
        return query


class MWM_YSO_Variable_BOSS_Carton(BaseCarton):
    """YSOs - Variable BOSS (pre-main sequence optical variables).

    Shorthand name: mwm_yso_variable_boss

    old class name: MWM_YSO_S3_Carton
    old shorthand name: mwm_yso_s3

    Simplified Description of selection criteria:
    selection of YSOs brighter than H<13, closer than parallax>0.3.
    Filter on the position of the HR diagram to
    select cool pre-main sequence stars,
    with BP-RP>13, (BP-RP)*2.5+2.5>M_G, (BP-RP)*2.5-1<M_G,
    requiring variability in g,bp,rp>0.02
    (with var_x defined as
    sqrt(phot_x_n_obs)/phot_x_mean_flux_over_error),
    have relations in variability of
    var_g<var_bp<var_g^0.75, 0.75*var_g<var_rp<var_g^0.95,
    and log10(var_bp)*5+11<M_BP, in which M_x is the absolute mag
    (should have ~52.7K sources)
    Wiki page:
    https://wiki.sdss.org/display/MWM/YSO+selection+function
    Additional source catalogs needed: 2mass, gaia
    Additional cross-matching needed:
    Return columns: Gaia id, 2mass id, G, BP, RP, J, H, K, parallax
    cadence options for these targets
    (list all options,
    even though no single target will receive more than one):
    boss_bright_3x1 if RP<14.76 |
    boss_bright_4x1 if RP<15.075 |
    boss_bright_5x1 if RP<15.29 |
    boss_bright_6x1 if RP<15.5
    Pseudo SQL (optional):
    Implementation:
    phot_rp_mean_mag<15.5 and phot_g_mean_mag < 18.5 and h_m <13 and parallax >0.3 and
    bp_rp*2.5+2.5 > phot_g_mean_mag-5*(log10(1000/parallax)-1) and
    bp_rp*2.5-1 < phot_g_mean_mag-5*(log10(1000/parallax)-1) and
    sqrt(phot_bp_n_obs)/phot_bp_mean_flux_over_error>
    sqrt(phot_g_n_obs)/phot_g_mean_flux_over_error and
    sqrt(phot_rp_n_obs)/phot_rp_mean_flux_over_error>
    sqrt(phot_g_n_obs)/phot_g_mean_flux_over_error*0.75 and
    sqrt(phot_bp_n_obs)/phot_bp_mean_flux_over_error<
    power(sqrt(phot_g_n_obs)/phot_g_mean_flux_over_error,0.75) and
    sqrt(phot_rp_n_obs)/phot_rp_mean_flux_over_error<
    power(sqrt(phot_g_n_obs)/phot_g_mean_flux_over_error,0.95) and
    log10(sqrt(phot_bp_n_obs)/phot_bp_mean_flux_over_error)*5+11<
    phot_bp_mean_mag-5*(log10(1000/parallax)-1) and
    bp_rp>1.3 and sqrt(phot_g_n_obs)/phot_g_mean_flux_over_error>0.02
    and
    sqrt(phot_bp_n_obs)/phot_bp_mean_flux_over_error>0.02 and
    sqrt(phot_rp_n_obs)/phot_rp_mean_flux_over_error>0.02

    Comments: Split from mwm_yso_s3 to request BOSS observations,
    RP magnitude check added to the previous selection
    """

    name = 'mwm_yso_variable_boss'
    category = 'science'
    # cadence is set in post_process()
    cadence = None
    program = 'mwm_yso'
    mapper = 'MWM'
    priority = 2700

    def build_query(self, version_id, query_region=None):

        query = (CatalogToTIC_v8
                 .select(CatalogToTIC_v8.catalogid, Gaia_DR2.source_id,
                         Gaia_DR2.ra.alias('gaia_dr2_ra'),
                         Gaia_DR2.dec.alias('gaia_dr2_dec'),
                         TwoMassPSC.pts_key,
                         TwoMassPSC.designation.alias('twomass_psc_designation'),
                         Gaia_DR2.phot_g_mean_mag, Gaia_DR2.phot_bp_mean_mag,
                         Gaia_DR2.phot_rp_mean_mag.alias('gaia_dr2_rp'),
                         TwoMassPSC.j_m, TwoMassPSC.h_m,
                         TwoMassPSC.k_m, Gaia_DR2.parallax)
                 .join(TIC_v8, on=(CatalogToTIC_v8.target_id == TIC_v8.id))
                 .join(TwoMassPSC, on=(TIC_v8.twomass_psc == TwoMassPSC.designation))
                 .switch(TIC_v8)
                 .join(Gaia_DR2, on=(TIC_v8.gaia_int == Gaia_DR2.source_id))
                 .where(CatalogToTIC_v8.version_id == version_id,
                        CatalogToTIC_v8.best >> True,
                        Gaia_DR2.phot_rp_mean_mag < 15.5,
                        Gaia_DR2.phot_g_mean_mag < 18.5,
                        TwoMassPSC.h_m < 13,
                        Gaia_DR2.parallax > 0.3,
                        Gaia_DR2.bp_rp * 2.5 + 2.5 >
                        Gaia_DR2.phot_g_mean_mag -
                        5 * (peewee.fn.log(1000 / Gaia_DR2.parallax) - 1),
                        Gaia_DR2.bp_rp * 2.5 - 1 <
                        Gaia_DR2.phot_g_mean_mag -
                        5 * (peewee.fn.log(1000 / Gaia_DR2.parallax) - 1),
                        peewee.fn.sqrt(Gaia_DR2.phot_bp_n_obs) /
                        Gaia_DR2.phot_bp_mean_flux_over_error >
                        peewee.fn.sqrt(Gaia_DR2.phot_g_n_obs) /
                        Gaia_DR2.phot_g_mean_flux_over_error,
                        peewee.fn.sqrt(Gaia_DR2.phot_rp_n_obs) /
                        Gaia_DR2.phot_rp_mean_flux_over_error >
                        peewee.fn.sqrt(Gaia_DR2.phot_g_n_obs) /
                        Gaia_DR2.phot_g_mean_flux_over_error * 0.75,
                        peewee.fn.sqrt(Gaia_DR2.phot_bp_n_obs) /
                        Gaia_DR2.phot_bp_mean_flux_over_error <
                        peewee.fn.power(
                            peewee.fn.sqrt(Gaia_DR2.phot_g_n_obs) /
                            Gaia_DR2.phot_g_mean_flux_over_error, 0.75),
                        peewee.fn.sqrt(Gaia_DR2.phot_rp_n_obs) /
                        Gaia_DR2.phot_rp_mean_flux_over_error <
                        peewee.fn.power(
                            peewee.fn.sqrt(Gaia_DR2.phot_g_n_obs) /
                            Gaia_DR2.phot_g_mean_flux_over_error, 0.95),
                        peewee.fn.log(
                            peewee.fn.sqrt(Gaia_DR2.phot_bp_n_obs) /
                            Gaia_DR2.phot_bp_mean_flux_over_error) * 5 + 11 <
                        Gaia_DR2.phot_bp_mean_mag -
                        5 * (peewee.fn.log(1000 / Gaia_DR2.parallax) - 1),
                        Gaia_DR2.bp_rp > 1.3,
                        peewee.fn.sqrt(Gaia_DR2.phot_g_n_obs) /
                        Gaia_DR2.phot_g_mean_flux_over_error > 0.02,
                        peewee.fn.sqrt(Gaia_DR2.phot_bp_n_obs) /
                        Gaia_DR2.phot_bp_mean_flux_over_error > 0.02,
                        peewee.fn.sqrt(Gaia_DR2.phot_rp_n_obs) /
                        Gaia_DR2.phot_rp_mean_flux_over_error > 0.02))

        if query_region:
            query = (query
                     .join_from(CatalogToTIC_v8, Catalog)
                     .where(peewee.fn.q3c_radial_query(Catalog.ra,
                                                       Catalog.dec,
                                                       query_region[0],
                                                       query_region[1],
                                                       query_region[2])))
        return query

    def post_process(self, model):
        """
        cadence options for these targets:
        boss_bright_3x1 if RP<14.76 |
        boss_bright_4x1 if RP<15.075 |
        boss_bright_5x1 if RP<15.29 |
        boss_bright_6x1 if RP<15.5
        """

        cursor = self.database.execute_sql(
            "select catalogid, gaia_dr2_rp from " +
            " sandbox.temp_mwm_yso_variable_boss ;")

        output = cursor.fetchall()

        for i in range(len(output)):
            current_catalogid = output[i][0]
            current_rp = output[i][1]

            if(current_rp < 14.76):
                current_cadence = 'boss_bright_3x1'
            elif(current_rp < 15.075):
                current_cadence = 'boss_bright_4x1'
            elif(current_rp < 15.29):
                current_cadence = 'boss_bright_5x1'
            elif(current_rp < 15.5):
                current_cadence = 'boss_bright_6x1'
            else:
                current_cadence = None

            self.database.execute_sql(
                " update sandbox.temp_mwm_yso_variable_boss " +
                " set cadence = '" + current_cadence + "'"
                " where catalogid = " + str(current_catalogid) + ";")


class MWM_YSO_OB_APOGEE_Carton(BaseCarton):
    """YSOs - OB APOGEE Upper (pre-)Main Sequence.

    Shorthand name: mwm_yso_ob_apogee

    old class name: MWM_YSO_OB_Carton
    old shorthand name: mwm_yso_ob

    Simplified Description of selection criteria:
    Selecting the OB stars at the tip of the main sequence,
    brighter than H<13, G<18 mag, closer than parallax>0.3,
    color -0.2<BP-RP<1.1, and M_G<(BP-RP)*1.6-2.2
    (should have ~8.7K sources)
    Wiki page:
    https://wiki.sdss.org/display/MWM/YSO+selection+function
    Additional source catalogs needed: 2mass, gaia
    Additional cross-matching needed:
    Return columns: Gaia id, 2mass id, G, BP, RP, J, H, K, parallax
    cadence options for these targets
    (list all options,
    even though no single target will receive more than one):
    Pseudo SQL (optional):
    Implementation: h_m<13 and bp_rp between -0.2 and 1.1 and
    phot_g_mean_mag<18 and
    phot_g_mean_mag-5*(log10(1000/parallax)-1) <
    1.6*bp_rp-2.2 and parallax>0.3

    """

    name = 'mwm_yso_ob_apogee'
    category = 'science'
    cadence = None  # 'apogee_bright_3x1'
    program = 'mwm_yso'
    mapper = 'MWM'
    priority = 2700

    def build_query(self, version_id, query_region=None):

        query = (CatalogToTIC_v8
                 .select(CatalogToTIC_v8.catalogid, Gaia_DR2.source_id,
                         Gaia_DR2.ra.alias('gaia_dr2_ra'),
                         Gaia_DR2.dec.alias('gaia_dr2_dec'),
                         TwoMassPSC.pts_key,
                         TwoMassPSC.designation.alias('twomass_psc_designation'),
                         Gaia_DR2.phot_g_mean_mag, Gaia_DR2.phot_bp_mean_mag,
                         Gaia_DR2.phot_rp_mean_mag.alias('gaia_dr2_rp'),
                         TwoMassPSC.j_m, TwoMassPSC.h_m,
                         TwoMassPSC.k_m, Gaia_DR2.parallax)
                 .join(TIC_v8, on=(CatalogToTIC_v8.target_id == TIC_v8.id))
                 .join(TwoMassPSC, on=(TIC_v8.twomass_psc == TwoMassPSC.designation))
                 .switch(TIC_v8)
                 .join(Gaia_DR2, on=(TIC_v8.gaia_int == Gaia_DR2.source_id))
                 .where(CatalogToTIC_v8.version_id == version_id,
                        CatalogToTIC_v8.best >> True,
                        TwoMassPSC.h_m < 13,
                        (Gaia_DR2.bp_rp > -0.2) & (Gaia_DR2.bp_rp < 1.1),
                        Gaia_DR2.phot_g_mean_mag < 18,
                        Gaia_DR2.phot_g_mean_mag -
                        5 * (peewee.fn.log(1000 / Gaia_DR2.parallax) - 1) <
                        1.6 * Gaia_DR2.bp_rp - 2.2,
                        Gaia_DR2.parallax > 0.3))

        if query_region:
            query = (query
                     .join_from(CatalogToTIC_v8, Catalog)
                     .where(peewee.fn.q3c_radial_query(Catalog.ra,
                                                       Catalog.dec,
                                                       query_region[0],
                                                       query_region[1],
                                                       query_region[2])))

        return query


class MWM_YSO_OB_BOSS_Carton(BaseCarton):
    """YSOs - OB BOSS Upper (pre-)Main Sequence.

    Shorthand name: mwm_yso_ob_boss

    old class name: MWM_YSO_OB_Carton
    old shorthand name: mwm_yso_ob

    Simplified Description of selection criteria:
    Selecting the OB stars at the tip of the main sequence,
    brighter than rp<15.5, G<18 mag, closer than parallax>0.3,
    color -0.2<BP-RP<1.1, and M_G<(BP-RP)*1.6-2.2
    (should have ~8.7K sources)
    Wiki page:
    https://wiki.sdss.org/display/MWM/YSO+selection+function
    Additional source catalogs needed: 2mass, gaia
    Additional cross-matching needed:
    Return columns: Gaia id, 2mass id, G, BP, RP, J, H, K, parallax
    cadence options for these targets
    (list all options,
    even though no single target will receive more than one):
    boss_bright_3x1 if RP<14.76 |
    boss_bright_4x1 if RP<15.075 |
    boss_bright_5x1 if RP<15.29 |
    boss_bright_6x1 if RP<15.5
    Pseudo SQL (optional):
    Implementation: rp<15.5 and bp_rp between -0.2 and 1.1 and
    phot_g_mean_mag<18 and
    phot_g_mean_mag-5*(log10(1000/parallax)-1) <
    1.6*bp_rp-2.2 and parallax>0.3

    Comments: Split from mwm_yso_ob to request BOSS observations,
    assigning cadence and faint limit for carton based on RP instead of H
    """

    name = 'mwm_yso_ob_boss'
    category = 'science'
    # cadence is set in post_process()
    cadence = None
    program = 'mwm_yso'
    mapper = 'MWM'
    priority = 2700

    def build_query(self, version_id, query_region=None):

        query = (CatalogToTIC_v8
                 .select(CatalogToTIC_v8.catalogid, Gaia_DR2.source_id,
                         Gaia_DR2.ra.alias('gaia_dr2_ra'),
                         Gaia_DR2.dec.alias('gaia_dr2_dec'),
                         TwoMassPSC.pts_key,
                         TwoMassPSC.designation.alias('twomass_psc_designation'),
                         Gaia_DR2.phot_g_mean_mag, Gaia_DR2.phot_bp_mean_mag,
                         Gaia_DR2.phot_rp_mean_mag.alias('gaia_dr2_rp'),
                         TwoMassPSC.j_m, TwoMassPSC.h_m,
                         TwoMassPSC.k_m, Gaia_DR2.parallax)
                 .join(TIC_v8, on=(CatalogToTIC_v8.target_id == TIC_v8.id))
                 .join(TwoMassPSC, on=(TIC_v8.twomass_psc == TwoMassPSC.designation))
                 .switch(TIC_v8)
                 .join(Gaia_DR2, on=(TIC_v8.gaia_int == Gaia_DR2.source_id))
                 .where(CatalogToTIC_v8.version_id == version_id,
                        CatalogToTIC_v8.best >> True,
                        Gaia_DR2.phot_rp_mean_mag < 15.5,
                        (Gaia_DR2.bp_rp > -0.2) & (Gaia_DR2.bp_rp < 1.1),
                        Gaia_DR2.phot_g_mean_mag < 18,
                        Gaia_DR2.phot_g_mean_mag -
                        5 * (peewee.fn.log(1000 / Gaia_DR2.parallax) - 1) <
                        1.6 * Gaia_DR2.bp_rp - 2.2,
                        Gaia_DR2.parallax > 0.3))

        if query_region:
            query = (query
                     .join_from(CatalogToTIC_v8, Catalog)
                     .where(peewee.fn.q3c_radial_query(Catalog.ra,
                                                       Catalog.dec,
                                                       query_region[0],
                                                       query_region[1],
                                                       query_region[2])))

        return query

    def post_process(self, model):
        """
        cadence options for these targets:
        boss_bright_3x1 if RP<14.76 |
        boss_bright_4x1 if RP<15.075 |
        boss_bright_5x1 if RP<15.29 |
        boss_bright_6x1 if RP<15.5
        """

        cursor = self.database.execute_sql(
            "select catalogid, gaia_dr2_rp from " +
            " sandbox.temp_mwm_yso_ob_boss ;")

        output = cursor.fetchall()

        for i in range(len(output)):
            current_catalogid = output[i][0]
            current_rp = output[i][1]

            if(current_rp < 14.76):
                current_cadence = 'boss_bright_3x1'
            elif(current_rp < 15.075):
                current_cadence = 'boss_bright_4x1'
            elif(current_rp < 15.29):
                current_cadence = 'boss_bright_5x1'
            elif(current_rp < 15.5):
                current_cadence = 'boss_bright_6x1'
            else:
                current_cadence = None

            self.database.execute_sql(
                " update sandbox.temp_mwm_yso_ob_boss " +
                " set cadence = '" + current_cadence + "'"
                " where catalogid = " + str(current_catalogid) + ";")


class MWM_YSO_CMZ_APOGEE_Carton(BaseCarton):
    """YSOs - Central Molecular Zone APOGEE.

    Shorthand name: mwm_yso_cmz_apogee

    old class name: MWM_YSO_CMZ_Carton
    old shorthand name: mwm_yso_cmz

    Simplified Description of selection criteria:
    selection of sources in the central molecular zone
    based on spitzer fluxes from mipsgal.
     brighter than H<13, have color 8.0-24>2.5, and
     have parallax<0.2 or lack a Gaia xmatch.
    (should have ~3.2K sources)
    Wiki page:
    https://wiki.sdss.org/display/MWM/YSO+selection+function
    Additional source catalogs needed: mipsgal
    Additional cross-matching needed: the table has xmatch included
    Return columns:
    mipsgal id, 2mass id, j, h, k, 3.6, 4.8, 8.0, 24 mag
    cadence options for these targets
    (list all options,
    even though no single target will receive more than one):
    'apogee_bright_3x1'
    Pseudo SQL (optional):
    Implementation: Hmag<13 and _8_0_-_24_>2.5 and
    (parallax<0.2 or parallax is null)

    For CMZ, the raw sql query would be:
    select ct.catalogid from mipsgal m
    join twomass_psc t on twomass_name = designation
    join tic_v8 tic on tic.twomass_psc = t.designation
    left outer join gaia_dr2_source g on g.source_id = tic.gaia_int
    join catalog_to_tic_v8 ct on ct.target_id = tic.id
    where m.hmag < 13 and
    (m.mag_8_0 - m.mag_24) > 2.5 and
    (g.parallax < 0.2 or g.parallax is null)
    and ct.version_id = 13 and ct.best is true;

    Note you only need one left outer join between TIC and Gaia
    (all MIPSGAL targets have a counterpart in 2MASS,
    and all 2MASS have an entry in TIC,
    but not all the TIC entries have a Gaia counterpart).

    Comments: Formerly mwm_yso_cmz, removed check on the position on the sky:

    Removed below condition.
    l is glon (galactic longitude)
    b is glat (galactic latitude)
    All four statements below are equivalent.
    (l> 358 or l< 2) and
    b between -1 and 1

    (m.glon > 358 or m.glon < 2) and
    (m.glat > -1 and m.glat < 1) and

    Sources are within 2 degrees in l and
     1 degree in b from the galactic center,

    (MIPSGAL.glon > 358) | (MIPSGAL.glon < 2),
    (MIPSGAL.glat > -1) & (MIPSGAL.glat < 1),

    """

    name = 'mwm_yso_cmz_apogee'
    category = 'science'
    cadence = None  # 'apogee_bright_3x1'
    program = 'mwm_yso'
    mapper = 'MWM'
    priority = 2700

    # mipsgal is a subset of 2MASS
    # mipsgal can be joined to twomass_psc via
    # mipsgal.twomass_name = TwoMassPSC.designation.
    # Then join via TIC and catalog_to_tic.
    #
    # mipsgal is a subset of 2MASS
    # 2MASS is a subset of TIC_v8
    # Gaia_DR2 is a subset of TIC_v8
    #
    # 2MASS is not a subset of Gaia_DR2
    # Gaia_DR2 is not a subset of 2MASS
    #
    # table catalogdb.mipsgal
    # Foreign-key constraints:
    #    "twomass_name_fk" FOREIGN KEY (twomass_name)
    # REFERENCES twomass_psc(designation)
    #
    # Due to below, we do not need a between to Catalog and CatalogToTIC_v8
    # Catalog.catalogid == CatalogToTIC_v8.catalogid
    # We can remove the join with Catalog in all the cartons
    # since catalogid is completely unique (even across different version_id)
    # so the join with Catalog doesn't give us anything extra and it's a costly join.

    def build_query(self, version_id, query_region=None):

        query = (MIPSGAL.select(CatalogToTIC_v8.catalogid, Gaia_DR2.source_id,
                                Gaia_DR2.ra.alias('gaia_dr2_ra'),
                                Gaia_DR2.dec.alias('gaia_dr2_dec'),
                                TwoMassPSC.pts_key,
                                TwoMassPSC.designation.alias('twomass_psc_designation'),
                                TwoMassPSC.j_m, TwoMassPSC.h_m,
                                TwoMassPSC.k_m, MIPSGAL.mag_3_6, MIPSGAL.mag_4_5,
                                MIPSGAL.mag_5_8, MIPSGAL.mag_8_0, MIPSGAL.mag_24,
                                MIPSGAL.hmag, Gaia_DR2.parallax)
                 .join(TwoMassPSC, on=(MIPSGAL.twomass_name == TwoMassPSC.designation))
                 .join(TIC_v8, on=(TIC_v8.twomass_psc == TwoMassPSC.designation))
                 .join(Gaia_DR2, peewee.JOIN.LEFT_OUTER,
                       on=(Gaia_DR2.source_id == TIC_v8.gaia_int))
                 .switch(TIC_v8)
                 .join(CatalogToTIC_v8, on=(CatalogToTIC_v8.target_id == TIC_v8.id))
                 .where(CatalogToTIC_v8.version_id == version_id,
                        CatalogToTIC_v8.best >> True,
                        MIPSGAL.hmag < 13,
                        (MIPSGAL.mag_8_0 - MIPSGAL.mag_24) > 2.5,
                        (Gaia_DR2.parallax < 0.2) |
                        (Gaia_DR2.parallax >> None)))

        if query_region:
            query = (query
                     .join_from(CatalogToTIC_v8, Catalog)
                     .where(peewee.fn.q3c_radial_query(Catalog.ra,
                                                       Catalog.dec,
                                                       query_region[0],
                                                       query_region[1],
                                                       query_region[2])))

        return query


class MWM_YSO_Cluster_APOGEE_Carton(BaseCarton):
    """YSOs - Cluster APOGEE Catalog
    Shorthand name: mwm_yso_cluster_apogee

    old class name: MWM_YSO_Cluster_Carton
    old shorthand name: mwm_yso_cluster

    Simplified Description of selection criteria:
    Selecting the clustered sources from
    the catalog of clustered structures,
    with age<7.5 dex and brighter than H<13 mag.
    (should have ~45.5K sources)
    Wiki page:
    https://wiki.sdss.org/display/MWM/YSO+selection+function
    Additional source catalogs needed: Kounkel+20 clustered catalog
    Additional cross-matching needed:
    Return columns: Gaia id, 2mass id, G, BP, RP, J, H, K, parallax
    cadence options for these targets
    (list all options,
    even though no single target will receive more than one):
    Pseudo SQL (optional):
    Implementation: age<7.5 and h<13

    """

    name = 'mwm_yso_cluster_apogee'
    category = 'science'
    cadence = None  # 'apogee_bright_3x1'
    program = 'mwm_yso'
    mapper = 'MWM'
    priority = 2700

    # yso_clustering is a subset of gaia and
    # can be joined to gaia_dr2_source via source_id.
    #
    # table catalogdb.yso_clustering
    # Foreign-key constraints:
    #    "yso_clustering_source_id_fkey" FOREIGN KEY (source_id)
    # REFERENCES gaia_dr2_source(source_id)

    def build_query(self, version_id, query_region=None):

        query = (CatalogToTIC_v8
                 .select(CatalogToTIC_v8.catalogid, Gaia_DR2.source_id,
                         Gaia_DR2.ra.alias('gaia_dr2_ra'),
                         Gaia_DR2.dec.alias('gaia_dr2_dec'),
                         YSO_Clustering.twomass,
                         Gaia_DR2.phot_g_mean_mag, Gaia_DR2.phot_bp_mean_mag,
                         Gaia_DR2.phot_rp_mean_mag.alias('gaia_dr2_rp'),
                         YSO_Clustering.j, YSO_Clustering.h,
                         YSO_Clustering.k, Gaia_DR2.parallax)
                 .join(TIC_v8, on=(CatalogToTIC_v8.target_id == TIC_v8.id))
                 .join(Gaia_DR2, on=(TIC_v8.gaia_int == Gaia_DR2.source_id))
                 .join(YSO_Clustering,
                       on=(Gaia_DR2.source_id == YSO_Clustering.source_id))
                 .where(CatalogToTIC_v8.version_id == version_id,
                        CatalogToTIC_v8.best >> True,
                        YSO_Clustering.h < 13,
                        YSO_Clustering.age < 7.5))

        if query_region:
            query = (query
                     .join_from(CatalogToTIC_v8, Catalog)
                     .where(peewee.fn.q3c_radial_query(Catalog.ra,
                                                       Catalog.dec,
                                                       query_region[0],
                                                       query_region[1],
                                                       query_region[2])))

        return query


class MWM_YSO_Cluster_BOSS_Carton(BaseCarton):
    """YSOs - Cluster BOSS Catalog
    Shorthand name: mwm_yso_cluster_boss

    old class name: MWM_YSO_Cluster_Carton
    old shorthand name: mwm_yso_cluster

    Simplified Description of selection criteria:
    Selecting the clustered sources from
    the catalog of clustered structures,
    with age<7.5 dex and brighter than rp<15.5 mag.

    Wiki page:
    https://wiki.sdss.org/display/MWM/YSO+selection+function
    Additional source catalogs needed: Kounkel+20 clustered catalog
    Additional cross-matching needed:
    Return columns: Gaia id, 2mass id, G, BP, RP, J, H, K, parallax
    cadence options for these targets
    (list all options,
    even though no single target will receive more than one):
    cadence options for these targets:
    boss_bright_3x1 if RP<14.76 |
    boss_bright_4x1 if RP<15.075 |
    boss_bright_5x1 if RP<15.29 |
    boss_bright_6x1 if RP<15.5
    Pseudo SQL (optional):
    Implementation: age<7.5 and rp<15.5

    Comments: Split from Cluster to request BOSS observations,
    assigning cadence and faint limit for carton based on RP instead of H
    """

    name = 'mwm_yso_cluster_boss'
    category = 'science'
    # cadence is assigned in post_process()
    cadence = None
    program = 'mwm_yso'
    mapper = 'MWM'
    priority = 2700

    # yso_clustering is a subset of gaia and
    # can be joined to gaia_dr2_source via source_id.
    #
    # table catalogdb.yso_clustering
    # Foreign-key constraints:
    #    "yso_clustering_source_id_fkey" FOREIGN KEY (source_id)
    # REFERENCES gaia_dr2_source(source_id)

    def build_query(self, version_id, query_region=None):

        query = (CatalogToTIC_v8
                 .select(CatalogToTIC_v8.catalogid, Gaia_DR2.source_id,
                         Gaia_DR2.ra.alias('gaia_dr2_ra'),
                         Gaia_DR2.dec.alias('gaia_dr2_dec'),
                         YSO_Clustering.twomass,
                         Gaia_DR2.phot_g_mean_mag, Gaia_DR2.phot_bp_mean_mag,
                         Gaia_DR2.phot_rp_mean_mag.alias('gaia_dr2_rp'),
                         YSO_Clustering.j, YSO_Clustering.h,
                         YSO_Clustering.k, Gaia_DR2.parallax)
                 .join(TIC_v8, on=(CatalogToTIC_v8.target_id == TIC_v8.id))
                 .join(Gaia_DR2, on=(TIC_v8.gaia_int == Gaia_DR2.source_id))
                 .join(YSO_Clustering,
                       on=(Gaia_DR2.source_id == YSO_Clustering.source_id))
                 .where(CatalogToTIC_v8.version_id == version_id,
                        CatalogToTIC_v8.best >> True,
                        Gaia_DR2.phot_rp_mean_mag < 15.5,
                        YSO_Clustering.age < 7.5))

        if query_region:
            query = (query
                     .join_from(CatalogToTIC_v8, Catalog)
                     .where(peewee.fn.q3c_radial_query(Catalog.ra,
                                                       Catalog.dec,
                                                       query_region[0],
                                                       query_region[1],
                                                       query_region[2])))

        return query

    def post_process(self, model):
        """
        cadence options for these targets:
        boss_bright_3x1 if RP<14.76 |
        boss_bright_4x1 if RP<15.075 |
        boss_bright_5x1 if RP<15.29 |
        boss_bright_6x1 if RP<15.5
        """

        cursor = self.database.execute_sql(
            "select catalogid, gaia_dr2_rp from " +
            " sandbox.temp_mwm_yso_cluster_boss ;")

        output = cursor.fetchall()

        for i in range(len(output)):
            current_catalogid = output[i][0]
            current_rp = output[i][1]

            if(current_rp < 14.76):
                current_cadence = 'boss_bright_3x1'
            elif(current_rp < 15.075):
                current_cadence = 'boss_bright_4x1'
            elif(current_rp < 15.29):
                current_cadence = 'boss_bright_5x1'
            elif(current_rp < 15.5):
                current_cadence = 'boss_bright_6x1'
            else:
                current_cadence = None

            self.database.execute_sql(
                " update sandbox.temp_mwm_yso_cluster_boss " +
                " set cadence = '" + current_cadence + "'"
                " where catalogid = " + str(current_catalogid) + ";")


class MWM_YSO_PMS_APOGEE_Carton(BaseCarton):
    """
    YSOs - Pre-main sequence, APOGEE
    Shorthand name: mwm_yso_pms_apogee
    Comments: New
    Simplified Description of selection criteria:
    Selecting the clustered sources from the catalog of vetted
    pre-main sequence stars
    Wiki page: https://wiki.sdss.org/display/MWM/YSO+selection+function
    Additional source catalogs needed: catalogdb.sagitta, catalogdb.zari18pms
    Return columns: Gaia id, 2mass id, G, BP, RP, J, H, K, parallax
    cadence options for these targets
    (list all options, even though no single target will receive more than one):
    apogee_bright_3x1 (for 7 < H < 13)
    Implementation: (in sagitta | in zari18pms) & h<13
    lead contact:Marina Kounkel
    """

    # peewee Model name ---> postgres table name
    # Gaia_DR2(CatalogdbModel)--->'gaia_dr2_source'
    # Zari18pms(CatalogdbModel)--->'catalogdb.zari18pms'
    # Zari18ums(CatalogdbModel)--->'catalogdb.zari18ums'
    # Sagitta(CatalogdbModel)--->'catalogdb.sagitta'
    # TwoMassPSC(CatalogdbModel)--->'catalogdb.twomass_psc'

    name = 'mwm_yso_pms_apogee'
    category = 'science'
    cadence = None  # 'apogee_bright_3x1'
    program = 'mwm_yso'
    mapper = 'MWM'
    priority = 2700

    def build_query(self, version_id, query_region=None):

        # join with Sagitta
        query1 = (CatalogToTIC_v8
                  .select(CatalogToTIC_v8.catalogid, Gaia_DR2.source_id,
                          Gaia_DR2.ra.alias('gaia_dr2_ra'),
                          Gaia_DR2.dec.alias('gaia_dr2_dec'),
                          TwoMassPSC.pts_key,
                          TwoMassPSC.designation.alias('twomass_psc_designation'),
                          Gaia_DR2.phot_g_mean_mag, Gaia_DR2.phot_bp_mean_mag,
                          Gaia_DR2.phot_rp_mean_mag.alias('gaia_dr2_rp'),
                          TwoMassPSC.j_m, TwoMassPSC.h_m,
                          TwoMassPSC.k_m, Gaia_DR2.parallax)
                  .join(TIC_v8, on=(CatalogToTIC_v8.target_id == TIC_v8.id))
                  .join(Gaia_DR2, on=(TIC_v8.gaia_int == Gaia_DR2.source_id))
                  .switch(TIC_v8)
                  .join(TwoMassPSC, on=(TIC_v8.twomass_psc == TwoMassPSC.designation))
                  .switch(Gaia_DR2)
                  .join(Sagitta,
                        on=(Gaia_DR2.source_id == Sagitta.source_id))
                  .where(CatalogToTIC_v8.version_id == version_id,
                         CatalogToTIC_v8.best >> True,
                         TwoMassPSC.h_m < 13,
                         TwoMassPSC.h_m > 7))

        # join with Zari18pms
        query2 = (CatalogToTIC_v8
                  .select(CatalogToTIC_v8.catalogid, Gaia_DR2.source_id,
                          Gaia_DR2.ra.alias('gaia_dr2_ra'),
                          Gaia_DR2.dec.alias('gaia_dr2_dec'),
                          TwoMassPSC.pts_key,
                          TwoMassPSC.designation.alias('twomass_psc_designation'),
                          Gaia_DR2.phot_g_mean_mag, Gaia_DR2.phot_bp_mean_mag,
                          Gaia_DR2.phot_rp_mean_mag.alias('gaia_dr2_rp'),
                          TwoMassPSC.j_m, TwoMassPSC.h_m,
                          TwoMassPSC.k_m, Gaia_DR2.parallax)
                  .join(TIC_v8, on=(CatalogToTIC_v8.target_id == TIC_v8.id))
                  .join(Gaia_DR2, on=(TIC_v8.gaia_int == Gaia_DR2.source_id))
                  .switch(TIC_v8)
                  .join(TwoMassPSC, on=(TIC_v8.twomass_psc == TwoMassPSC.designation))
                  .switch(Gaia_DR2)
                  .join(Zari18pms,
                        on=(Gaia_DR2.source_id == Zari18pms.source))
                  .where(CatalogToTIC_v8.version_id == version_id,
                         CatalogToTIC_v8.best >> True,
                         TwoMassPSC.h_m < 13,
                         TwoMassPSC.h_m > 7))

        # | is for peewee SQL union
        query = query1 | query2

        if query_region:
            query = (query
                     .join_from(CatalogToTIC_v8, Catalog)
                     .where(peewee.fn.q3c_radial_query(Catalog.ra,
                                                       Catalog.dec,
                                                       query_region[0],
                                                       query_region[1],
                                                       query_region[2])))

        return query


class MWM_YSO_PMS_BOSS_Carton(BaseCarton):
    """
    YSOs - Pre-main sequence, BOSS
    Shorthand name: mwm_yso_pms_boss
    Comments: New, Split from PMS
    Simplified Description of selection criteria:
    Selecting the clustered sources from the catalog of vetted
    pre-main sequence stars
    Wiki page: https://wiki.sdss.org/display/MWM/YSO+selection+function
    Additional source catalogs needed: catalogdb.sagitta, catalogdb.zari18pms
    Return columns: Gaia id, 2mass id, G, BP, RP, J, H, K, parallax
    cadence options for these targets:
    boss_bright_3x1 if RP<14.76 |
    boss_bright_4x1 if RP<15.075 |
    boss_bright_5x1 if RP<15.29 |
    boss_bright_6x1 if RP<15.5
    Implementation: (in sagitta | in zari18pms) & rp<15.5
    lead contact:Marina Kounkel
    """

    # peewee Model name ---> postgres table name
    # Gaia_DR2(CatalogdbModel)--->'gaia_dr2_source'
    # Zari18pms(CatalogdbModel)--->'catalogdb.zari18pms'
    # Zari18ums(CatalogdbModel)--->'catalogdb.zari18ums'
    # Sagitta(CatalogdbModel)--->'catalogdb.sagitta'
    # TwoMassPSC(CatalogdbModel)--->'catalogdb.twomass_psc'

    name = 'mwm_yso_pms_boss'
    category = 'science'
    # cadence is assigned in post_process()
    cadence = None
    program = 'mwm_yso'
    mapper = 'MWM'
    priority = 2700

    def build_query(self, version_id, query_region=None):

        # join with Sagitta
        query1 = (CatalogToTIC_v8
                  .select(CatalogToTIC_v8.catalogid, Gaia_DR2.source_id,
                          Gaia_DR2.ra.alias('gaia_dr2_ra'),
                          Gaia_DR2.dec.alias('gaia_dr2_dec'),
                          TwoMassPSC.pts_key,
                          TwoMassPSC.designation.alias('twomass_psc_designation'),
                          Gaia_DR2.phot_g_mean_mag, Gaia_DR2.phot_bp_mean_mag,
                          Gaia_DR2.phot_rp_mean_mag.alias('gaia_dr2_rp'),
                          TwoMassPSC.j_m, TwoMassPSC.h_m,
                          TwoMassPSC.k_m, Gaia_DR2.parallax)
                  .join(TIC_v8, on=(CatalogToTIC_v8.target_id == TIC_v8.id))
                  .join(Gaia_DR2, on=(TIC_v8.gaia_int == Gaia_DR2.source_id))
                  .switch(TIC_v8)
                  .join(TwoMassPSC, on=(TIC_v8.twomass_psc == TwoMassPSC.designation))
                  .switch(Gaia_DR2)
                  .join(Sagitta,
                        on=(Gaia_DR2.source_id == Sagitta.source_id))
                  .where(CatalogToTIC_v8.version_id == version_id,
                         CatalogToTIC_v8.best >> True,
                         Gaia_DR2.phot_rp_mean_mag < 15.5,
                         Gaia_DR2.phot_rp_mean_mag > 7))

        # join with Zari18pms
        query2 = (CatalogToTIC_v8
                  .select(CatalogToTIC_v8.catalogid, Gaia_DR2.source_id,
                          Gaia_DR2.ra.alias('gaia_dr2_ra'),
                          Gaia_DR2.dec.alias('gaia_dr2_dec'),
                          TwoMassPSC.pts_key,
                          TwoMassPSC.designation.alias('twomass_psc_designation'),
                          Gaia_DR2.phot_g_mean_mag, Gaia_DR2.phot_bp_mean_mag,
                          Gaia_DR2.phot_rp_mean_mag.alias('gaia_dr2_rp'),
                          TwoMassPSC.j_m, TwoMassPSC.h_m,
                          TwoMassPSC.k_m, Gaia_DR2.parallax)
                  .join(TIC_v8, on=(CatalogToTIC_v8.target_id == TIC_v8.id))
                  .join(Gaia_DR2, on=(TIC_v8.gaia_int == Gaia_DR2.source_id))
                  .switch(TIC_v8)
                  .join(TwoMassPSC, on=(TIC_v8.twomass_psc == TwoMassPSC.designation))
                  .switch(Gaia_DR2)
                  .join(Zari18pms,
                        on=(Gaia_DR2.source_id == Zari18pms.source))
                  .where(CatalogToTIC_v8.version_id == version_id,
                         CatalogToTIC_v8.best >> True,
                         Gaia_DR2.phot_rp_mean_mag < 15.5,
                         Gaia_DR2.phot_rp_mean_mag > 7))

        # | is for peewee SQL union
        query = query1 | query2

        if query_region:
            query = (query
                     .join_from(CatalogToTIC_v8, Catalog)
                     .where(peewee.fn.q3c_radial_query(Catalog.ra,
                                                       Catalog.dec,
                                                       query_region[0],
                                                       query_region[1],
                                                       query_region[2])))

        return query

    def post_process(self, model):
        """
        cadence options for these targets:
        boss_bright_3x1 if RP<14.76 |
        boss_bright_4x1 if RP<15.075 |
        boss_bright_5x1 if RP<15.29 |
        boss_bright_6x1 if RP<15.5
        """

        cursor = self.database.execute_sql(
            "select catalogid, gaia_dr2_rp from " +
            " sandbox.temp_mwm_yso_boss_pms ;")

        output = cursor.fetchall()

        for i in range(len(output)):
            current_catalogid = output[i][0]
            current_rp = output[i][1]

            if(current_rp < 14.76):
                current_cadence = 'boss_bright_3x1'
            elif(current_rp < 15.075):
                current_cadence = 'boss_bright_4x1'
            elif(current_rp < 15.29):
                current_cadence = 'boss_bright_5x1'
            elif(current_rp < 15.5):
                current_cadence = 'boss_bright_6x1'
            else:
                current_cadence = None

            self.database.execute_sql(
                " update sandbox.temp_mwm_yso_boss_pms " +
                " set cadence = '" + current_cadence + "'"
                " where catalogid = " + str(current_catalogid) + ";")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2020-04-11
# @Filename: utils.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import contextlib
import hashlib
import io
import time

import click
import pandas
from peewee import SQL, DoesNotExist, fn


__all__ = ['Timer', 'sql_apply_pm', 'sql_iauname', 'copy_pandas',
           'get_epoch', 'set_config_parameter', 'remove_version',
           'vacuum_outputs', 'get_configuration_values', 'vacuum_table']


class Timer:
    """Convenience context manager to time events.

    Modified from https://bit.ly/3ebdp3y.

    """

    def __enter__(self):
        self.start = time.time()
        self.end = None
        return self

    def __exit__(self, *args):
        self.end = time.time()
        self.interval = self.end - self.start

    @property
    def elapsed(self):

        if self.end:
            return self.interval

        return time.time() - self.start


def sql_apply_pm(ra_field, dec_field, pmra_field, pmdec_field,
                 epoch_delta, is_pmra_cos=True):
    """Constructs a SQL expression for applying proper motions to RA/Dec.

    Parameters
    ----------
    ra_field
        The Peewee field or value representing Right Ascension.
    dec_field
        The Peewee field or value representing Declination.
    pmra_field
        The Peewee field or value representing the proper motion in RA.
    pmdec_field
        The Peewee field or value representing the proper motion in Dec.
    epoch_delta
        A value or expression with the delta epoch, in years.
    is_pmra_cos : bool
        Whether the ``pmra_field`` includes the correction for the cosine of
        the declination

    Returns
    -------
    ra_dec : `~peewee:Expression`
        A tuple with the `~peewee:Expression` instances for the proper motion
        corrected RA and Dec.

    """

    pmra_field = fn.coalesce(pmra_field, 0) / 1000. / 3600.
    pmdec_field = fn.coalesce(pmdec_field, 0) / 1000. / 3600.
    epoch = fn.coalesce(epoch_delta, 0)

    if is_pmra_cos:
        cos_dec = fn.cos(fn.radians(fn.coalesce(dec_field, 0)))
        ra_corr = ra_field + pmra_field / cos_dec * epoch
    else:
        ra_corr = ra_field + pmra_field * epoch

    dec_corr = dec_field + pmdec_field * epoch

    return ra_corr, dec_corr


def sql_iauname(ra_field, dec_field, prefix='SDSS J'):
    """Constructs a SQL expression for the IAU name from RA/Dec.

    Parameters
    ----------
    ra_field
        The Peewee field or value representing Right Ascension.
    dec_field
        The Peewee field or value representing Declination.
    prefix : str
        The prefix to add before the calculated IAU name.

    Returns
    -------
    iauname : `~peewee:Expression`
        The `~peewee:Expression` instance to construct the IAU name.

    """

    def dd2dms(dd, round=2, sign=False):

        degs = fn.trunc(dd).cast('INT')
        mins = fn.trunc((fn.abs(dd) - fn.abs(degs)) * 60).cast('INT')
        secs = (fn.round(((((fn.abs(dd) - fn.abs(degs)) * 60) - mins) * 60)
                         .cast('NUMERIC'), round).cast('REAL'))

        degs_fmt = '00' if sign is False else 'SG00'
        secs_fmt = '00.' + '0' * round

        return fn.trim(fn.to_char(degs, degs_fmt)).concat(
            fn.trim(fn.to_char(mins, '00'))).concat(
                fn.trim(fn.to_char(secs, secs_fmt)))

    return SQL('\'' + prefix + '\'').concat(
        dd2dms(ra_field / 15)).concat(
            dd2dms(dec_field, round=1, sign=True))


def copy_pandas(df, database, table_name, schema=None, columns=None):
    """Inserts a Pandas data frame using COPY.

    Parameters
    ----------
    df : ~pandas.DataFrame
        The Pandas data frame to copy or the inputs to be used to create one.
    database
        A database connection.
    table_name : str
        The name of the table into which to copy the data.
    schema : str
        The schema in which the table lives.
    columns : list
        The list of column names to copy.

    """

    if not isinstance(df, pandas.DataFrame):
        df = pandas.DataFrame(data=df)

    stream = io.StringIO()
    df.to_csv(stream, index=False, header=False)
    stream.seek(0)

    cursor = database.cursor()

    full_table_name = schema + '.' + table_name if schema else table_name

    with database.atomic():
        cursor.copy_from(stream, full_table_name, columns=columns, sep=',')

    return df


def get_epoch(xmodel):
    """Returns the epoch for an `.XMatchModel` in Julian years."""

    xmatch = xmodel._meta.xmatch
    fields = xmodel._meta.fields

    if not xmatch.epoch and not xmatch.epoch_column:
        return None

    # If epoch == 0, make it null. This helps with q3c functions.
    if xmatch.epoch:
        epoch = xmatch.epoch
    else:
        epoch = fn.nullif(fields[xmatch.epoch_column], 0)

    if xmatch.epoch_format == 'jd':
        epoch = 2000 + (epoch - 2451545.0) / 365.25

    return epoch


@contextlib.contextmanager
def set_config_parameter(database, parameter, new_value, reset=True, log=None):
    """Temporarily a database configuration parameter."""

    new_value = new_value.upper()

    try:

        orig_value = (database
                      .execute_sql(f'SHOW {parameter}')
                      .fetchone()[0].upper())
        value_changed = orig_value != new_value

        if value_changed:
            database.execute_sql(f'SET {parameter} = {new_value};')
            if log:
                log.debug(f'{parameter} is {new_value}.')

        yield

    finally:

        if reset and value_changed:
            database.execute_sql(f'SET {parameter} = {orig_value};')
            if log:
                log.debug(f'{parameter} reset to {orig_value}.')


def remove_version(database, plan, schema='catalogdb', table='catalog',
                   delete_version=True, vacuum=True):
    """Removes all rows in ``table`` and ``table_to_`` that match a version."""

    models = []
    for path in database.models:
        schema, table_name = path.split('.')
        if schema != schema:
            continue
        if table_name != table and not table_name.startswith(table + '_to_'):
            continue
        model = database.models[path]
        if not model.table_exists() or model._meta.schema != schema:
            continue
        models.append(model)

    if len(models) == 0:
        raise ValueError('No table matches the input parameters.')

    print(f'Tables that will be truncated on {plan!r}: ' +
          ', '.join(model._meta.table_name for model in models))

    Catalog = database.models[schema + '.' + table]
    Version = database.models[schema + '.version']

    try:
        version_id = Version.get(plan=plan).id
    except DoesNotExist:
        raise ValueError(f'Version {plan!r} does not exist.')

    print(f'version_id={version_id}')

    n_targets = (Catalog
                 .select()
                 .where(Catalog.version_id == version_id)
                 .count())
    print(f'Number of targets found in '
          f'{Catalog._meta.table_name}: {n_targets:,}')

    if not click.confirm('Do you really want to proceed?'):
        return

    for model in models:
        n_removed = (model
                     .delete()
                     .where(model.version_id == version_id)
                     .execute())
        print(f'{model._meta.table_name}: {n_removed:,} rows removed.')
        if vacuum:
            print('Vacuuming ...')
            vacuum_table(database,
                         f'{model._meta.schema}.{model._meta.table_name}')

    md5 = hashlib.md5(plan.encode()).hexdigest()[0:16]
    for table in database.get_tables(schema=schema):
        if table.endswith(md5):
            print(f'Dropping temporary table {table}.')
            database.execute_sql(f'DROP TABLE IF EXISTS {table};')

    if delete_version:
        Version.delete().where(Version.id == version_id).execute()
        print('Removed entry in \'version\'.')


def vacuum_table(database, table_name, vacuum=True, analyze=True,
                 maintenance_work_mem='50GB'):
    """Vacuums and analyses a table."""

    statement = (('VACUUM ' if vacuum else '') +
                 ('ANALYZE ' if analyze else '') +
                 table_name)

    with database.atomic():

        # Change isolation level to allow executing commands such as VACUUM.
        connection = database.connection()
        original_isolation_level = connection.isolation_level
        connection.set_isolation_level(0)

        database.execute_sql(
            f'SET maintenance_work_mem = {maintenance_work_mem!r}')

        database.execute_sql(statement)

        connection.set_isolation_level(original_isolation_level)


def vacuum_outputs(database, vacuum=True, analyze=True, schema='catalogdb',
                   table='catalog', relational_tables=True):
    """Vacuums and analyses the output tables."""

    assert vacuum or analyze, 'either vacuum or analyze need to be True.'
    assert database.is_connection_usable(), 'connection is not usable.'

    tables = []
    for full_name in database.models:
        table_schema, table_name = full_name.split('.')
        if table_schema != schema:
            continue
        if table_name != table:
            is_relational = table_name.startswith(table + '_to_')
            if not is_relational or not relational_tables:
                continue
        tables.append(table_name)

    for table_name in tables:
        table_name = (table_name
                      if schema is None
                      else schema + '.' + table_name)
        vacuum_table(database, table_name, vacuum=vacuum, analyze=analyze)


def get_configuration_values(database, parameters):
    """Returns a dictionary of datbase configuration parameter to value."""

    values = {}

    with database.atomic():
        for parameter in parameters:
            value = database.execute_sql(f'SHOW {parameter}').fetchone()[0]
            values[parameter] = value

    return values


def is_view(database, view_name, schema='public', materialized=False):
    """Determines if a view exists."""

    if not materialized:
        query = (f'SELECT * FROM pg_views where schemaname = {schema!r} '
                 f'AND viewname = {view_name!r};')
    else:
        query = (f'SELECT pc.* FROM pg_class pc '
                 f'JOIN pg_namespace pn ON pc.relnamespace = pn.oid '
                 f'WHERE pn.nspname = {schema!r} AND '
                 f'pc.relname = {view_name!r};')

    if len(database.execute_sql(query).fetchall()) > 0:
        return True
    else:
        return False



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: Tom Dwelly
# @Date: 2020-06-12
# @Filename: bhm_csc.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

# flake8: noqa
# isort: skip_file

import peewee
import sdssdb

from sdssdb.peewee.sdss5db.catalogdb import (Catalog,
                                             BHM_CSC,
                                             CatalogToBHM_CSC)


from target_selection.cartons.base import BaseCarton


# Details: Start here
# https://wiki.sdss.org/display/OPS/Defining+target+selection+and+cadence+algorithms#Definingtargetselectionandcadencealgorithms-ChandraSourceCatalogue(CSC)-inprogress

## This module provides the following BHM cartons:
# bhm_csc_boss-dark
# bhm_csc_boss-bright
# bhm_csc_apogee



class BhmCscBaseCarton(BaseCarton):
    '''
    Parent class that provides the basic selections for all CSC cartons
    To be sub-classed, not typically to be called directly.
    '''

    name = 'bhm_csc-base'
    base_name = 'bhm_csc-base'
    category = 'science'
    mapper = 'BHM'
    program = 'bhm_csc'
    tile = False
    priority = None
    alias_t = None
    instrument = None

    def build_query(self, version_id, query_region=None):
        c = Catalog.alias()
        c2t = CatalogToBHM_CSC.alias()
        t = BHM_CSC.alias()
        self.alias_t = t

        # set the Carton priority+values here - read from yaml
        priority = peewee.Value(int(self.parameters.get('priority', 10000))).alias('priority')
        value = peewee.Value(self.parameters.get('value', 1.0)).cast('float').alias('value')

        query = (
            c
            .select(c.catalogid,
                    priority,
                    value,
                    t.mag_g.alias('g'),
                    t.mag_r.alias('r'),
                    t.mag_i.alias('i'),
                    t.mag_z.alias('z'),
                    t.mag_h.alias('h'),
            )
            .join(c2t)
            .join(t)
            .where(c.version_id == version_id,
                   c2t.version_id == version_id,
                   c2t.best == True)
            .distinct([c2t.target_id])  # avoid duplicates - initially trust the CSC parent sample,
            .where
            (
                t.spectrograph == self.instrument
            )
        )

        return query


#-------bhm_csc_boss_dark ------ #

class BhmCscBossDarkCarton(BhmCscBaseCarton):
    '''
        SELECT * from bhm_csc AS c WHERE c.spectrograph = "BOSS" AND WHERE c.mag_i BETWEEN 17.x AND 21.x
    '''
    name = 'bhm_csc_boss-dark'
    cadence = 'bhm_csc_boss_1x4'
    instrument = 'BOSS'


    def build_query(self, version_id, query_region=None):
        query = super().build_query(version_id, query_region)
        t = self.alias_t
        query = query.where(
            (t.mag_i >= self.parameters['mag_i_min']) &
            (t.mag_i <  self.parameters['mag_i_max'])
        )
        return query

#-------bhm_csc_boss_bright ------ #

class BhmCscBossBrightCarton(BhmCscBaseCarton):
    '''
        SELECT * from bhm_csc AS c WHERE c.spectrograph = "BOSS" AND WHERE c.mag_i BETWEEN 1x.0 AND 18.x
    '''
    name = 'bhm_csc_boss-bright'
    cadence = 'bhm_csc_boss_1x1'
    instrument = 'BOSS'

    def build_query(self, version_id, query_region=None):
        query = super().build_query(version_id, query_region)
        t = self.alias_t
        query = query.where(
            (t.mag_i >= self.parameters['mag_i_min']) &
            (t.mag_i <  self.parameters['mag_i_max'])
        )
        return query

#-------bhm_csc_apogee ------ #

class BhmCscApogeeCarton(BhmCscBaseCarton):
    '''
        SELECT * from bhm_csc AS c WHERE c.spectrograph = "APOGEE" AND WHERE c.mag_H BETWEEN 1x.x AND 1x.x
    '''
    name = 'bhm_csc_apogee'
    cadence = 'bhm_csc_apogee_3x1'
    instrument = 'APOGEE'

    def build_query(self, version_id, query_region=None):
        query = super().build_query(version_id, query_region)
        t = self.alias_t
        query = query.where(
            (t.mag_h >= self.parameters['mag_h_min']) &
            (t.mag_h <  self.parameters['mag_h_max'])
        )
        return query

'''
Exporting from the temp table

\copy (SELECT * FROM sandbox.temp_bhm_csc_boss_dark)  TO '/home/tdwelly/scratch/targetdb/bhm_csc_boss_dark.csv' with csv header
\copy (SELECT * FROM sandbox.temp_bhm_csc_boss_bright)  TO '/home/tdwelly/scratch/targetdb/bhm_csc_boss_bright.csv' with csv header
\copy (SELECT * FROM sandbox.temp_bhm_csc_apogee)  TO '/home/tdwelly/scratch/targetdb/bhm_csc_apogee.csv' with csv header

for F in bhm_csc_*.csv; do   stilts tpipe in=${F} out="${F%.*}.fits" ifmt=csv ofmt=fits-basic; done
m
'''



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2020-07-26
# @Filename: ops_skies.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import peewee

from sdssdb.peewee.sdss5db.catalogdb import CatalogToSkies_v1, Skies_v1

from . import BaseCarton


class OPS_BOSS_Sky_Carton(BaseCarton):
    """Skies for the BOSS spectrograph.

    Definition:
        Select sky positions from catalogdb.skies_v1 that don't have a
        nearby Gaia, LS8, Tycho2, or 2MASS source.

    """

    name = 'ops_sky_boss'
    cadence = None
    category = 'ops_sky'
    program = 'SKY'
    mapper = None
    priority = 5000

    load_magnitudes = False

    def build_query(self, version_id, query_region):

        min_separation = 10 + ((12. - Skies_v1.mag_neighbour_gaia) / 0.2)

        query = (Skies_v1
                 .select(CatalogToSkies_v1.catalogid,
                         Skies_v1.ra,
                         Skies_v1.dec,
                         Skies_v1.pix_32768,
                         Skies_v1.tile_32)
                 .join(CatalogToSkies_v1)
                 .where(Skies_v1.sep_neighbour_gaia > min_separation)
                 .where(CatalogToSkies_v1.version_id == version_id,
                        CatalogToSkies_v1.best >> True)
                 .where(Skies_v1.gaia_sky >> True,
                        Skies_v1.ls8_sky >> True,
                        Skies_v1.tmass_sky >> True,
                        Skies_v1.tycho2_sky >> True,
                        Skies_v1.tmass_xsc_sky >> True))

        if query_region:
            query = (query
                     .where(peewee.fn.q3c_radial_query(Skies_v1.ra,
                                                       Skies_v1.dec,
                                                       query_region[0],
                                                       query_region[1],
                                                       query_region[2])))

        return query


class OPS_APOGEE_Sky_Carton(BaseCarton):
    """Skies for the APOGEE spectrograph.

    Definition:
        Select sky positions from catalogdb.skies_v1 that don't have a
        nearby Gaia, Tycho2, or 2MASS source.

    """

    name = 'ops_sky_apogee'
    cadence = None
    category = 'ops_sky'
    program = 'SKY'
    mapper = None
    priority = 5200

    load_magnitudes = False

    def build_query(self, version_id, query_region):

        query = (Skies_v1
                 .select(CatalogToSkies_v1.catalogid,
                         Skies_v1.ra, Skies_v1.dec,
                         Skies_v1.pix_32768, Skies_v1.tile_32)
                 .join(CatalogToSkies_v1)
                 .where(CatalogToSkies_v1.version_id == version_id,
                        CatalogToSkies_v1.best >> True)
                 .where(Skies_v1.gaia_sky >> True,
                        Skies_v1.tmass_sky >> True,
                        Skies_v1.tycho2_sky >> True,
                        Skies_v1.tmass_xsc_sky >> True))

        if query_region:
            query = (query
                     .where(peewee.fn.q3c_radial_query(Skies_v1.ra,
                                                       Skies_v1.dec,
                                                       query_region[0],
                                                       query_region[1],
                                                       query_region[2])))

        return query



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: Pramod Gupta (psgupta@uw.edu)
# @Date: 2020-10-16
# @Filename: mwm_legacy.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import peewee

from sdssdb.peewee.sdss5db.catalogdb import (Catalog, CatalogToTIC_v8,
                                             Gaia_DR2,
                                             SDSS_APOGEE_AllStarMerge_r13,
                                             TIC_v8, TwoMassPSC)

from target_selection.cartons import BaseCarton


# See catalog.py for the name of peewee model names corresponding
# to postgres table names:
# https://github.com/sdss/sdssdb/blob/master/python/sdssdb/peewee/sdss5db/catalogdb.py
#
# peewee model : postgres table name
# Catalog : catalogdb.catalog
# CatalogToTIC_v8 : catalogdb.catalog_to_tic_v8
# Gaia_DR2 : catalogdb.gaia_dr2_source
# SDSS_APOGEE_AllStarMerge_r13 : catalogdb.sdss_apogeeallstarmerge_r13
# TIC_v8 : catalogdb.tic_v8
# TwoMassPSC : catalogdb.twomass_psc
#

# We associate each APOGEE target with a 2MASS source since the
# vast majority of APOGEE targets were selected from the 2MASS PSC.
# APOGEE_ID is essentially the same as
# the “designation” column of the 2MASS PSC;
# We just have to strip off the “2M” in the APOGEE_ID.
#
# For example:
# sdss5db=# select designation from  catalogdb.twomass_psc limit 2;
#    designation
# ------------------
#  12034281-5740002
#  12032366-5738380
# (2 rows)
#
# sdss5db=# select apogee_id from  catalogdb.sdss_apogeeallstarmerge_r13 limit 2;
#      apogee_id
# --------------------
#  2M14044120-1550575
#  2M14033676-1554164
# (2 rows)
#
# sdss5db=# select ltrim(apogee_id,'2M') from
#  catalogdb.sdss_apogeeallstarmerge_r13 limit 2;
#       ltrim
# ------------------
#  14044120-1550575
#  14033676-1554164
# (2 rows)
#
#
# sdss5db=# select count(1) from
#  catalogdb.sdss_apogeeallstarmerge_r13 where dist_src='gaia';
#  count
# -------
#      0
# (1 row)
#
# Hence use trim(dist_src) like below:
#
# sdss5db=# select count(1) from
#  catalogdb.sdss_apogeeallstarmerge_r13 where trim(dist_src)='gaia';
#  count
# --------
#  487508
# (1 row)
#


class MWM_Legacy_ir2opt_Carton(BaseCarton):
    """MWM APOGEE targets.

Shorthand name: mwm_legacy_ir2opt

What is it?: APOGEE targets to be observed with BOSS fibres.
To be used as a filler for MWM-led plates.

Simplified Description of selection criteria: Select all Gaia targets that
have an APOGEE counterpart (i.e., have an entry in sdss_apogeeallstarmerge_r13)
with 13 < G < 18, BP > 13, and RP > 13.

Wiki page: NA

Additional source catalogs needed: Gaia, sdss_apogeeallstarmerge_r13

Additional cross-matching needed: NA

Return columns: catalog_id, source_id, apogee_id, phot_g_mean_mag,
phot_rp_mean_mag, phot_bp_mean_mag

cadence options for these targets
(list all options, even though no single target will receive more than one):
NA
    """

    name = 'mwm_legacy_ir2opt'
    category = 'science'
    cadence = None
    program = 'mwm_legacy'
    mapper = 'MWM'
    priority = 6100

    def build_query(self, version_id, query_region=None):

        query = (Catalog
                 .select(CatalogToTIC_v8.catalogid,
                         Gaia_DR2.source_id,
                         Gaia_DR2.phot_g_mean_mag,
                         Gaia_DR2.phot_bp_mean_mag,
                         Gaia_DR2.phot_rp_mean_mag,
                         TwoMassPSC.pts_key)
                 .join(CatalogToTIC_v8,
                       on=(Catalog.catalogid == CatalogToTIC_v8.catalogid))
                 .join(TIC_v8,
                       on=(CatalogToTIC_v8.target_id == TIC_v8.id))
                 .join(Gaia_DR2,
                       on=(TIC_v8.gaia_int == Gaia_DR2.source_id))
                 .switch(TIC_v8)
                 .join(TwoMassPSC,
                       on=(TIC_v8.twomass_psc == TwoMassPSC.designation))
                 .join(SDSS_APOGEE_AllStarMerge_r13,
                       on=(TwoMassPSC.designation ==
                           peewee.fn.ltrim(SDSS_APOGEE_AllStarMerge_r13.apogee_id,
                                           '2M')))
                 .where(CatalogToTIC_v8.version_id == version_id,
                        CatalogToTIC_v8.best >> True,
                        Gaia_DR2.phot_g_mean_mag.between(13, 18),
                        Gaia_DR2.phot_bp_mean_mag > 13,
                        Gaia_DR2.phot_rp_mean_mag > 13))

        # Gaia_DR2 peewee model class corresponds to
        # table catalogdb.gaia_dr2_source.

        if query_region:
            query = (query
                     .join_from(CatalogToTIC_v8, Catalog)
                     .where(peewee.fn.q3c_radial_query(Catalog.ra,
                                                       Catalog.dec,
                                                       query_region[0],
                                                       query_region[1],
                                                       query_region[2])))

        return query



#!/usr/bin/env python
# encoding: utf-8
#
# @Author: Tom Dwelly
# @Date: Oct-2019
# @Filename: mag_flux.py
# @License: BSD 3-Clause
# @Copyright: Tom Dwelly

# flake8: noqa
# isort: skip_file

from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

import numpy as np
from math import log10

__invalid_mag_val = 99.9999
__invalid_mag_thresh = 90.0
''' see http://www.sdss3.org/dr8/algorithms/magnitudes.php '''


# https://svn.sdss.org/public/repo/eboss/idlspec2d/trunk/pro/spec2d/readplugmap.pro
__psfflux_to_fiber2flux ={ "u" : (1./2.085),
                           "g" : (1./2.085),
                           "r" : (1./2.116),
                           "i" : (1./2.134),
                           "z" : (1./2.135),}



def fluxmag(flux, zp=None):
    assert zp is not None
    if np.ndim(flux) > 0:
        with np.errstate(divide='ignore', invalid='ignore'):
            return np.nan_to_num(zp - 2.5*np.log10(flux), copy=False,
                                 nan=__invalid_mag_val,
                                 posinf=__invalid_mag_val,
                                 neginf=__invalid_mag_val)
    else:
        if flux <= 0.0:
            return __invalid_mag_val
        else:
            return zp - 2.5*log10(flux)



def mag2flux(mag, zp=None):
    assert zp is not None
    if np.ndim(mag) > 0:
        with np.errstate(divide='ignore', invalid='ignore'):
            flux = np.where(mag >= __invalid_mag_thresh, 0.0, 10.0**(-0.4*(mag - zp)))
    else:
        if mag >= __invalid_mag_thresh:
            flux = 0.0
        else:
            flux = 10.0**(-0.4*(mag-zp))
    return flux


def psfflux2fiber2flux(psfflux, filt=None):
    try:
       c = __psfflux_to_fiber2flux[filt]
    except:
        raise Exception ( f"Unknown filter: {filt}" )
    return psfflux * c

def fiber2flux2psfflux(fiber2flux, filt=None):
    try:
       c = __psfflux_to_fiber2flux[filt]
    except:
        raise Exception ( f"Unknown filter: {filt}" )
    return fiber2flux / c

def psfmag2fiber2mag(psfmag, filt=None):
    try:
       c = __psfflux_to_fiber2flux[filt]
    except:
        raise Exception ( f"Unknown filter: {filt}" )
    return   psfmag - (2.5 * log10(c))


def fiber2mag2psfmag(fiber2mag, filt=None):
    try:
       c = __psfflux_to_fiber2flux[filt]
    except:
        raise Exception ( f"Unknown filter: {filt}" )
    return fiber2mag + (2.5 * log10(c))

def psfmag_minus_fiber2mag(filt=None):
    try:
       c = __psfflux_to_fiber2flux[filt]
    except:
        raise Exception ( f"Unknown filter: {filt}" )
    return   2.5 * log10(c)


def fiber2mag_minus_psfmag(filt=None):
    try:
       c = __psfflux_to_fiber2flux[filt]
    except:
        raise Exception ( f"Unknown filter: {filt}" )
    return -2.5 * log10(c)








def nMgy2AB(flux):
    return flux2mag(mag, 22.50)

def Jy2AB(flux):
    return flux2mag(mag, 8.90)

def mJy2AB(flux):
    return flux2mag(mag, 16.40)

def uJy2AB(flux):
    return flux2mag(mag, 23.90)


def AB2nMgy(mag):
    return mag2flux(mag, 22.50)

def AB2Jy(mag):
    return mag2flux(mag, 8.90)

def AB2mJy(mag):
    return mag2flux(mag, 16.40)

def AB2uJy(mag):
    return mag2flux(mag, 23.90)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2020-02-26
# @Filename: guide.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import peewee

from sdssdb.peewee.sdss5db.catalogdb import (Catalog, CatalogToTIC_v8,
                                             Gaia_DR2, Gaia_DR2_Clean, TIC_v8)

from .base import BaseCarton


Gaia = Gaia_DR2
Gaia_Clean = Gaia_DR2_Clean


class GuideCarton(BaseCarton):

    name = 'guide'
    category = 'guide'
    cadence = None

    def build_query(self, version_id, query_region=None):

        sample = (Catalog
                  .select(Catalog.catalogid,
                          Catalog.ra,
                          Catalog.dec,
                          Gaia.phot_g_mean_mag)
                  .join(CatalogToTIC_v8)
                  .join(TIC_v8)
                  .join(Gaia)
                  .join(Gaia_Clean)
                  .where((Gaia.phot_g_mean_mag > self.parameters['g_min']) &
                         (Gaia.phot_g_mean_mag < self.parameters['g_max']))
                  .where(Catalog.version_id == version_id,
                         CatalogToTIC_v8.version_id == version_id))

        if query_region:
            sample = (sample
                      .where(peewee.fn.q3c_radial_query(Catalog.ra,
                                                        Catalog.dec,
                                                        query_region[0],
                                                        query_region[1],
                                                        query_region[2])))

        sample = sample.cte('sample')

        # We should use q3c_join_pm and q3c_dist_pm here.
        subq = (Gaia
                .select(Gaia.source_id)
                .where(peewee.fn.q3c_join(sample.c.ra, sample.c.dec,
                                          Gaia.ra,
                                          Gaia.dec,
                                          self.parameters['min_separation']))
                .order_by(peewee.fn.q3c_dist(sample.c.ra,
                                             sample.c.dec,
                                             Gaia.ra,
                                             Gaia.dec).asc())
                .limit(1)
                .offset(1))

        decollided = (sample
                      .select(sample.c.catalogid)
                      .join(subq, peewee.JOIN.LEFT_LATERAL,
                            on=peewee.SQL('true'))
                      .where(subq.c.source_id.is_null())
                      .with_cte(sample))

        return decollided
